/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002, 2004, 2005                                  */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.cron;

import com.ibm.staf.*;
import com.ibm.staf.service.*;
import java.util.*;
import java.io.*;
import org.python.util.PythonInterpreter;
import org.python.core.*;

public class CronService implements STAFServiceInterfaceLevel30, Runnable
{
    private STAFHandle fHandle;
    private final String kVersion = "3.0.0";

    // Version of STAF (or later) required for this service
    private final String kRequiredSTAFVersion = "3.0.0";

    private STAFCommandParser fParmsParser = new STAFCommandParser();
    private STAFCommandParser fRegisterParser;
    private STAFCommandParser fListParser;
    private STAFCommandParser fVersionParser;
    private STAFMapClassDefinition fCronIDMapClass;
    private STAFMapClassDefinition fCronIDLongMapClass;
    private Hashtable fRegTable;
    private static boolean DEBUG = true;
    private Thread fQueueThread;
    private STAFCommandParser fUnregisterParser;
    private int fCronID = 1;
    private String fLineSep;
    private String fCronServiceMachine = "";
    private String fServiceName = "";
    private String fJythonVersion = "";
    private PythonInterpreter fPythonInterpreter;
    private String fHashtableFileDirectory;
    private String fHashtableFileName;
    private String fLocalMachineName;
    private final String kMinute = "minute";
    private final String kHour = "hour";
    private final String kDay = "day";
    private final String kMonth = "month";
    private final String kWeekday = "weekday";
    private HashMap fSubmittedRequests = new HashMap();
    static final String sQueueTypeEnd = "STAF/Service/Cron/End";

    public CronService()
    {
    }

    public STAFResult init(STAFServiceInterfaceLevel30.InitInfo info)
    {
        int rc = STAFResult.Ok;

        try
        {
            STAFResult res = new STAFResult();
            fHandle = new STAFHandle("STAF/SERVICE/" + info.name);
            fServiceName = info.name;

            // Resolve the machine name variable for the local machine

            res = STAFUtil.resolveInitVar("{STAF/Config/Machine}", fHandle);

            if (res.rc != STAFResult.Ok) return res;

            fLocalMachineName = res.result;

            if (info.serviceJar != null)
            {
                // Create the Jython Library directory if it doesn't exist yet
                // from the jython library data in the service's jarfile.
                // Set the classpath to point to the service's jar file and to
                // the jython.jar.

                res = STAFServiceSharedJython.setupJython(
                    info.serviceJar, fHandle, info.writeLocation);

                if (res.rc != 0)
                {
                    System.out.println(res.result);
                    return res;
                }

                // Get the version of Jython distributed with the Cron service
                // Note the getJythonVersion method was added in STAF V3.0.0

                try
                {
                    fJythonVersion = STAFServiceSharedJython.getJythonVersion();
                }
                catch (Error err)
                {
                    return new STAFResult(
                        STAFResult.STAFRegistrationError,
                        fServiceName + " service V" + kVersion +
                        " requires STAF V" + kRequiredSTAFVersion +
                        " or later");
                }

                // Initialize the PythonInterpreter's python.home to point to the
                // shared jython installation.

                Properties p = new Properties();
                p.setProperty("python.home",
                              STAFServiceSharedJython.getJythonDirName());
                PythonInterpreter.initialize(System.getProperties(), p, null);
            }

            // instantiate parsers as not case sensitive
            fRegisterParser = new STAFCommandParser(0, false);
            fListParser = new STAFCommandParser(0, false);
            fUnregisterParser = new STAFCommandParser(0, false);
            fVersionParser = new STAFCommandParser(0, false);

            fRegisterParser.addOption("REGISTER", 1,
                                      STAFCommandParser.VALUENOTALLOWED);

            fRegisterParser.addOption("MACHINE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("PYTHONMACHINE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOptionGroup("MACHINE PYTHONMACHINE", 1, 1);

            fRegisterParser.addOption("SERVICE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("PYTHONSERVICE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOptionGroup("SERVICE PYTHONSERVICE", 1, 1);

            fRegisterParser.addOption("REQUEST", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("PYTHONREQUEST", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOptionGroup("REQUEST PYTHONREQUEST", 1, 1);

            fRegisterParser.addOption("PREPARE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("MINUTE", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("HOUR", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("DAY", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("MONTH", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("WEEKDAY", 1,
                                      STAFCommandParser.VALUEREQUIRED);

            // Require at least one time interval option to be specified
            fRegisterParser.addOptionGroup(
                "MINUTE HOUR DAY MONTH WEEKDAY", 1, 5);

            fUnregisterParser.addOption("UNREGISTER", 1,
                                        STAFCommandParser.VALUENOTALLOWED);

            fUnregisterParser.addOption("ID", 1,
                                        STAFCommandParser.VALUEREQUIRED);

            fUnregisterParser.addOptionNeed("UNREGISTER", "ID");

            fListParser.addOption("LIST", 1,
                                  STAFCommandParser.VALUENOTALLOWED);

            fListParser.addOption("MACHINE", 1,
                                  STAFCommandParser.VALUEREQUIRED);

            fListParser.addOption("LONG", 1,
                                  STAFCommandParser.VALUENOTALLOWED);

            fVersionParser.addOption("VERSION", 1,
                                     STAFCommandParser.VALUENOTALLOWED);

            fVersionParser.addOption("JYTHON", 1,
                                     STAFCommandParser.VALUENOTALLOWED);

            fVersionParser.addOptionGroup("VERSION", 1, 1);
            fVersionParser.addOptionNeed("JYTHON", "VERSION");

            fCronIDMapClass = new STAFMapClassDefinition(
                "STAF/Service/Cron/CronID");
            fCronIDMapClass.addKey("cronID", "ID");
            fCronIDMapClass.addKey("machine", "Machine");
            fCronIDMapClass.addKey("service", "Service");
            fCronIDMapClass.addKey("request", "Request");
            fCronIDMapClass.addKey("minute", "Minute");
            fCronIDMapClass.addKey("hour", "Hour");
            fCronIDMapClass.addKey("dayOfMonth", "Day of Month");
            fCronIDMapClass.setKeyProperty("dayOfMonth", "display-short-name",
                                           "Day");
            fCronIDMapClass.addKey("month", "Month");
            fCronIDMapClass.addKey("dayOfWeek", "Day of Week");
            fCronIDMapClass.setKeyProperty("dayOfWeek", "display-short-name",
                                           "Weekday");

            fCronIDLongMapClass = new STAFMapClassDefinition(
                "STAF/Service/Cron/CronIDLong");
            fCronIDLongMapClass.addKey("cronID", "ID");
            fCronIDLongMapClass.addKey("machine", "Machine");
            fCronIDLongMapClass.addKey("machineType", "Machine Type");
            fCronIDLongMapClass.addKey("service", "Service");
            fCronIDLongMapClass.addKey("serviceType", "Service Type");
            fCronIDLongMapClass.addKey("request", "Request");
            fCronIDLongMapClass.addKey("requestType","Request Type");
            fCronIDLongMapClass.addKey("prepareScript", "Prepare Script");
            fCronIDLongMapClass.addKey("minute", "Minute");
            fCronIDLongMapClass.addKey("hour", "Hour");
            fCronIDLongMapClass.addKey("dayOfMonth", "Day of Month");
            fCronIDLongMapClass.addKey("month", "Month");
            fCronIDLongMapClass.addKey("dayOfWeek", "Day of Week");

            // Resolve the line separator variable for the local machine

            res = STAFUtil.resolveInitVar("{STAF/Config/Sep/Line}", fHandle);

            if (res.rc != STAFResult.Ok) return res;

            fLineSep = res.result;

            // Resolve the file separator variable for the local machine

            res = STAFUtil.resolveInitVar("{STAF/Config/Sep/File}", fHandle);

            if (res.rc != STAFResult.Ok) return res;

            String fileSep = res.result;

            // Store data for the cron service in directory:
            //   <STAF writeLocation>/service/<cron service name (lower-case)>

            fHashtableFileDirectory = info.writeLocation;

            if (!fHashtableFileDirectory.endsWith(fileSep))
            {
                fHashtableFileDirectory += fileSep;
            }

            fHashtableFileDirectory = fHashtableFileDirectory +
                "service" + fileSep + fServiceName.toLowerCase();

            File dir = new File(fHashtableFileDirectory);

            if (!dir.exists())
            {
                dir.mkdirs();
            }

            fHashtableFileName = fHashtableFileDirectory + fileSep +
                "cron.ser";

            loadHashtable();

            fQueueThread = new Thread(this);
            fQueueThread.start(); // this calls the run() method
        }
        catch (STAFException e)
        {
            return new STAFResult(STAFResult.STAFRegistrationError,
                                  e.toString());
        }

        return new STAFResult(rc);
    }

    private STAFResult handleParms(STAFServiceInterfaceLevel30.InitInfo info)
        throws STAFException
    {
        return new STAFResult(STAFResult.Ok);
    }

    public STAFResult acceptRequest(STAFServiceInterfaceLevel30.RequestInfo info)
    {
        String lowerRequest = info.request.toLowerCase();

        // call the appropriate method to handle the command
        if (lowerRequest.startsWith("register"))
        {
            return handleRegister(info);
        }
        else if (lowerRequest.startsWith("unregister"))
        {
            return handleUnregister(info);
        }
        else if (lowerRequest.startsWith("list"))
        {
            return handleList(info);
        }
        else if (lowerRequest.startsWith("help"))
        {
            return handleHelp(info);
        }
        else if (lowerRequest.startsWith("version"))
        {
            return handleVersion(info);
        }
        else
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  "Unknown Cron Request: " +
                                  lowerRequest);
        }
    }

    private STAFResult handleHelp(STAFServiceInterfaceLevel30.RequestInfo info)
    {
        // Verify the requester has at least trust level 1

        STAFResult trustResult = STAFUtil.validateTrust(
            1, fServiceName, "HELP", fLocalMachineName, info);

        if (trustResult.rc != STAFResult.Ok) return trustResult;

        // Return help text

        return new STAFResult(STAFResult.Ok,
         "Cron Service Help" + fLineSep
         + "REGISTER   MACHINE <machine> | PYTHONMACHINE <machine>"
         + fLineSep +
         "           SERVICE <service> | PYTHONSERVICE <machine>"
         + fLineSep +
         "           REQUEST <request> | PYTHONREQUEST <request>"
         + fLineSep + "           [PREPARE <script>]"
         + fLineSep + "           [MINUTE <minute>] [HOUR <hour>] "
         + fLineSep + "           [DAY <day>] [MONTH <month>] "
         + fLineSep + "           [WEEKDAY <weekday>]"
         + fLineSep + "UNREGISTER ID <registrationID>"
         + fLineSep + "LIST       [MACHINE <machine>] [LONG]"
         + fLineSep + "VERSION    [JYTHON]" + fLineSep + "HELP");
    }

    private STAFResult handleList(STAFServiceInterfaceLevel30.RequestInfo info)
    {
        STAFCommandParseResult parsedRequest = fListParser.parse(info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(
                STAFResult.InvalidRequestString, parsedRequest.errorBuffer);
        }

        STAFMarshallingContext mc = new STAFMarshallingContext();
        List cronIDList = new ArrayList();

        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        String machineParm = "";
        String typeParm = "";

        try
        {
            // Verify the requester has at least trust level 2

            STAFResult trustResult = STAFUtil.validateTrust(
                2, fServiceName, "LIST", fLocalMachineName, info);

            if (trustResult.rc != STAFResult.Ok) return trustResult;

            // Process the list request

            int registrationTableSize = fRegTable.size();

            if (registrationTableSize > 0)
            {
                // Resolve the MACHINE option

                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("machine"),
                    fHandle, info.requestNumber);

                if (resolvedValue.rc != 0) return resolvedValue;

                machineParm = resolvedValue.result;

                // Check if the LONG option is specified

                boolean longFormat = false;

                if (parsedRequest.optionTimes("LONG") > 0)
                {
                    longFormat = true;
                    mc.setMapClassDefinition(fCronIDLongMapClass);
                }
                else
                {
                    mc.setMapClassDefinition(fCronIDMapClass);
                }

                CronData theData;
                boolean match;
                Enumeration registrationEnum = fRegTable.elements();
                Enumeration keys = fRegTable.keys();

                for (Enumeration e = fRegTable.elements();
                     e.hasMoreElements();)
                {
                    Integer id = (Integer) keys.nextElement();
                    match = false;
                    theData = (CronData) e.nextElement();

                    String minutes = "";
                    String hours = "";
                    String days = "";
                    String months = "";
                    String weekdays = "";

                    if (machineParm.equals("") |
                        machineParm.equals(theData.fMachine))
                    {
                        Enumeration minuteEnum = theData.fMinute.elements();

                        while (minuteEnum.hasMoreElements())
                        {
                            minutes += (minuteEnum.nextElement()).toString();

                            if (minuteEnum.hasMoreElements())
                                minutes += ",";
                        }

                        Enumeration hourEnum = theData.fHour.elements();

                        while (hourEnum.hasMoreElements())
                        {
                            hours += (hourEnum.nextElement()).toString();

                            if (hourEnum.hasMoreElements())
                                hours += ",";
                        }

                        Enumeration dayEnum = theData.fDay.elements();

                        while (dayEnum.hasMoreElements())
                        {
                            days += (dayEnum.nextElement()).toString();

                            if (dayEnum.hasMoreElements())
                                days += ",";
                        }

                        Enumeration monthEnum = theData.fMonth.elements();

                        while (monthEnum.hasMoreElements())
                        {
                            months += (monthEnum.nextElement()).toString();

                            if (monthEnum.hasMoreElements())
                                months += ",";
                        }

                        Enumeration weekdayEnum = theData.fWeekday.elements();

                        while (weekdayEnum.hasMoreElements())
                        {
                            weekdays += (weekdayEnum.nextElement()).toString();

                            if (weekdayEnum.hasMoreElements())
                                weekdays += ",";
                        }

                        // Create a map representing the matching registration
                        // and add it to the list of registrations

                        Map cronIDMap;

                        if (longFormat)
                            cronIDMap = fCronIDLongMapClass.createInstance();
                        else
                            cronIDMap = fCronIDMapClass.createInstance();

                        cronIDMap.put("cronID", id.toString());
                        cronIDMap.put("machine", theData.fMachine);
                        cronIDMap.put("service", theData.fService);
                        cronIDMap.put("request", theData.fRequest);
                        
                        if ((minutes != null) && (!minutes.equals("")))
                        {
                            cronIDMap.put("minute", minutes);
                        }
                        if ((hours != null) && (!hours.equals("")))
                        {
                            cronIDMap.put("hour", hours);
                        }
                        if ((days != null) && (!days.equals("")))
                        {
                            cronIDMap.put("dayOfMonth", days);
                        }
                        if ((months != null) && (!months.equals("")))
                        {
                            cronIDMap.put("month", months);
                        }
                        if ((weekdays != null) && (!weekdays.equals("")))
                        {
                            cronIDMap.put("dayOfWeek", weekdays);
                        }

                        if (longFormat)
                        {
                            if (theData.fPythonMachine)
                                cronIDMap.put("machineType", "Python");
                            else
                                cronIDMap.put("machineType", "Literal");

                            if (theData.fPythonService)
                                cronIDMap.put("serviceType", "Python");
                            else
                                cronIDMap.put("serviceType", "Literal");

                            if (theData.fPythonRequest)
                                cronIDMap.put("requestType", "Python");
                            else
                                cronIDMap.put("requestType", "Literal");

                            if ((theData.fPrepare != null) &&
                                (!theData.fPrepare.equals("")))
                            {
                                cronIDMap.put("prepareScript",
                                              theData.fPrepare);
                            }
                        }

                        cronIDList.add(cronIDMap);
                    }
                }
            }
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }
            return new STAFResult(STAFResult.JavaError,
                                  "Internal Java error.");
        }

        mc.setRootObject(cronIDList);

        return new STAFResult(STAFResult.Ok, mc.marshall());
    }

    private STAFResult handleRegister(
                           STAFServiceInterfaceLevel30.RequestInfo info)
    {
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        String originMachine = info.endpoint;
        String machine = "";
        boolean pythonMachine = true;
        String service = "";
        boolean pythonService = true;
        String request = "";
        boolean pythonRequest = true;
        String prepare = "";

        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        STAFCommandParseResult parsedRequest =
            fRegisterParser.parse(info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parsedRequest.errorBuffer);
        }

        // For each option, if the user has specified a value, set the
        // corresponding local variable to the user-specified value.

        try
        {
            // Verify the requester has at least trust level 5

            STAFResult trustResult = STAFUtil.validateTrust(
                5, fServiceName, "REGISTER", fLocalMachineName, info);

            if (trustResult.rc != STAFResult.Ok) return trustResult;

            // Process the request

            if (parsedRequest.optionTimes("machine") > 0)
            {
                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("machine"),
                    fHandle, info.requestNumber);

                if (resolvedValue.rc != 0) return resolvedValue;

                machine = resolvedValue.result;
                pythonMachine = false;
            }
            else if (parsedRequest.optionTimes("pythonmachine") > 0)
            {
                machine = parsedRequest.optionValue("pythonmachine");
                pythonMachine = true;
            }

            if (parsedRequest.optionTimes("service") > 0)
            {
                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("service"),
                    fHandle, info.requestNumber);

                if (resolvedValue.rc != 0) return resolvedValue;

                service = resolvedValue.result;
                pythonService = false;
            }
            else if (parsedRequest.optionTimes("pythonservice") > 0)
            {
                service = parsedRequest.optionValue("pythonservice");
                pythonService = true;
            }

            if (parsedRequest.optionTimes("request") > 0)
            {
                resolvedValue = STAFUtil.resolveRequestVar(
                    parsedRequest.optionValue("request"),
                    fHandle, info.requestNumber);

                if (resolvedValue.rc != 0) return resolvedValue;

                request = resolvedValue.result;
                pythonRequest = false;
            }
            else if (parsedRequest.optionTimes("pythonrequest") > 0)
            {
                request = parsedRequest.optionValue("pythonrequest");
                pythonRequest = true;
            }

            if (parsedRequest.optionTimes("prepare") > 0)
            {
                prepare = parsedRequest.optionValue("prepare");
            }

            Vector minute = new Vector();
            Vector hour = new Vector();
            Vector day = new Vector();
            Vector month = new Vector();
            Vector weekday = new Vector();

            if (parsedRequest.optionTimes(kMinute) > 0)
            {
                minute = cronParse(parsedRequest.optionValue(kMinute), kMinute);

                if (minute == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kMinute + ":  " +
                        parsedRequest.optionValue(kMinute));
                }
            }

            if (parsedRequest.optionTimes(kHour) > 0)
            {
                hour = cronParse(parsedRequest.optionValue(kHour), kHour);

                if (hour == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kHour + ":  " +
                        parsedRequest.optionValue(kHour));
                }
            }

            if (parsedRequest.optionTimes(kDay) > 0)
            {
                day = cronParse(parsedRequest.optionValue(kDay), kDay);

                if (day == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kDay + ":  " +
                        parsedRequest.optionValue(kDay));
                }
            }

            if (parsedRequest.optionTimes(kMonth) > 0)
            {
                month = cronParse(parsedRequest.optionValue(kMonth), kMonth);

                if (month == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kMonth + ":  " +
                        parsedRequest.optionValue(kMonth));
                }
            }

            if (parsedRequest.optionTimes(kWeekday) > 0)
            {
                weekday = cronParse(parsedRequest.optionValue(kWeekday),
                                    kWeekday);

                if (weekday == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kWeekday + ":  " +
                        parsedRequest.optionValue(kWeekday));
                }
            }

            Integer index = new Integer(fCronID++);

            fRegTable.put(
                index,
                (Object) new CronData(originMachine, machine, pythonMachine,
                                      service, pythonService, request,
                                      pythonRequest, prepare, minute, hour,
                                      day, month, weekday));

            sResult = new STAFResult(STAFResult.Ok, index.toString());
            saveHashtable();

            String message = "Register ID=" + index + " originMachine=" +
                originMachine + " machine=" + machine + " pythonMachine=" +
                pythonMachine + " service=" + service + " pythonService=" +
                pythonService + " request=" + request + " pythonRequest=" +
                pythonRequest + " prepare=" + prepare + " minute=" + minute +
                " hour=" + hour + " day=" + day + " month=" + month +
                " weekday=" + weekday;

            fHandle.submit2(
                "local", "LOG", "LOG MACHINE LOGNAME " +
                fServiceName + " LEVEL info MESSAGE " +
                STAFUtil.wrapData(message));
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }
            return new STAFResult(STAFResult.JavaError,
                                  "Internal Java error.");
        }

        return sResult;
    }

    private STAFResult handleUnregister(
        STAFServiceInterfaceLevel30.RequestInfo info)
    {
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        String service;
        String request;
        Integer id;

        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        STAFCommandParseResult parsedRequest =
            fUnregisterParser.parse(info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parsedRequest.errorBuffer);
        }

        // For each option, if the user has specified a value, set the
        // corresponding local variable to the user-specified value.

        try
        {
            // Verify the requester has at least trust level 4

            STAFResult trustResult = STAFUtil.validateTrust(
                4, fServiceName, "UNREGISTER", fLocalMachineName, info);

            if (trustResult.rc != STAFResult.Ok) return trustResult;

            // Process the request

            resolvedValue = STAFUtil.resolveRequestVar(
                parsedRequest.optionValue("id"), fHandle, info.requestNumber);

            if (resolvedValue.rc != 0) return resolvedValue;
            
            id = new Integer(resolvedValue.result);

            if (!fRegTable.containsKey(id))
            {
                return new STAFResult(STAFResult.DoesNotExist,
                                      "ID " + id + " not found");
            }
            else
            {
                CronData data = (CronData)fRegTable.get(id);

                fRegTable.remove(id);

                if (fRegTable.isEmpty())
                {
                    fCronID = 1;
                }

                saveHashtable();

                String message = "Unregister ID=" + id;

                fHandle.submit2(
                    "local", "LOG", "LOG MACHINE LOGNAME " +
                    fServiceName + " level info message " +
                    STAFUtil.wrapData(message));
            }
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }

            return new STAFResult(
                STAFResult.JavaError, "Internal Java error.");
        }

        return sResult;
    }

    private STAFResult handleVersion(
        STAFServiceInterfaceLevel30.RequestInfo info)
    {
        // Verify the requester has at least trust level 1

        STAFResult trustResult = STAFUtil.validateTrust(
            1, fServiceName, "VERSION", fLocalMachineName, info);

        if (trustResult.rc != STAFResult.Ok) return trustResult;

        // Parse the request

        STAFCommandParseResult parsedRequest = fVersionParser.parse(
            info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parsedRequest.errorBuffer);
        }

        if (parsedRequest.optionTimes("JYTHON") == 0)
        {
            // Return the version of the service
            return new STAFResult(STAFResult.Ok, kVersion);
        }
        else
        {
            // Return the version of Python packaged with the service
            return new STAFResult(STAFResult.Ok, fJythonVersion);
        }
    }

    private Vector cronParse(String msg, String type)
    {
        Vector result = new Vector();
        
        if (msg.equalsIgnoreCase("ANY") || msg.equals("*"))
        {
            result.add("ANY");
            return result;
        }
                
        StringTokenizer options = new StringTokenizer(msg, ",");
            
        while (options.hasMoreElements())
        {
            String element = ((String)options.nextElement()).trim();
        
            // Check if a range was specified:  <start>-<end>

            int firstDashIndex = element.indexOf("-");

            if (firstDashIndex == -1)
            {
                // Single value specified.  Verify that it is valid.
                    
                if (!isValidValue(element, type))
                    return null;

                result.add(element);
            }
            else
            {
                // Range was specified

                if (element.indexOf("-", firstDashIndex + 1) > -1)
                {
                    // Invalid.  Can only contain one dash.
                    return null;
                }

                StringTokenizer items = new StringTokenizer(element, "-");

                String start = null;
                String end = null;

                while (items.hasMoreElements())
                {
                    if (start == null)
                        start = ((String)items.nextElement()).trim();
                    else
                        end = ((String)items.nextElement()).trim();
                }

                // Verify that the values specified for the range are valid
                
                if (!isValidValue(start, type))
                    return null;

                if (!isValidValue(end, type))
                    return null;

                // Types other than weekday will always contain numeric values
                // A Weekday type can contain a numeric value or a string value

                if (!type.equals(kWeekday))
                {
                    // Type is kMonth, kDay, kHour, or kMinute

                    // Iterate through the numeric range

                    int startNum = new Integer(start).intValue();
                    int endNum = new Integer(end).intValue();

                    for (int i = startNum; i <= endNum; i++)
                    {
                        result.add(new String(new Integer(i).toString()));
                    }
                }
                else
                {
                    // Type is kWeekday

                    // Check if using string or numeric values for weekday.

                    boolean startIsNumeric = true;
                    boolean endIsNumeric = true;
                        
                    if (equalsWeekday(start)) startIsNumeric = false;
                    if (equalsWeekday(end)) endIsNumeric = false;
                        
                    int startNum = 0;
                    int endNum = 0;

                    // Get the corresponding numeric value for weekday if
                    // specified as a string, e.g. 0 for "Sunday".

                    if (startIsNumeric)
                        startNum = new Integer(start).intValue();
                    else
                        startNum = getNumericWeekday(start);

                    if (endIsNumeric)
                        endNum = new Integer(end).intValue();
                    else
                        endNum = getNumericWeekday(end);
                        
                    // Iterate throught the range, using the String or numeric
                    // representation based on the start value.

                    for (int i = startNum; i <= endNum; i++)
                    {
                        if (startIsNumeric)
                            result.add(new String(new Integer(i).toString()));
                        else
                            result.add(getStringWeekday(i));
                    }
                }
            }
        }

        return result;
    }
    
    boolean isValidValue(String value, String type)
    {
        if (value == null)
            return false;

        int valueNum = 0;

        if (type.equals(kWeekday))
        {
            if (equalsWeekday(value))
                return true;
        }
            
        try
        {
            valueNum = new Integer(value).intValue();
        }
        catch (NumberFormatException ex)
        {
            return false;
        }
        
        if (type.equals(kMinute))
        {
            if (valueNum >= 0 && valueNum <= 59)
                return true;
            else
                return false;
        }
        else if (type.equals(kHour))
        {
            if (valueNum >= 0 && valueNum <= 23)
                return true;
            else
                return false;
        }
        else if (type.equals(kDay))
        {
            if (valueNum >= 1 && valueNum <= 31)
                return true;
            else
                return false;
        }
        else if (type.equals(kMonth))
        {
            if (valueNum >= 1 && valueNum <= 12)
                return true;
            else
                return false;
        }
        else if (type.equals(kWeekday))
        {
            if (valueNum >= 0 && valueNum <= 6)
                return true;
            else
                return false;
        }
        else
        {
            return false;
        }
    }

    private boolean equalsWeekday(String msg)
    {
         if ((msg.equalsIgnoreCase("Sunday")) ||
             (msg.equalsIgnoreCase("Monday")) ||
             (msg.equalsIgnoreCase("Tuesday")) ||
             (msg.equalsIgnoreCase("Wednesday")) ||
             (msg.equalsIgnoreCase("Thursday")) ||
             (msg.equalsIgnoreCase("Friday")) ||
             (msg.equalsIgnoreCase("Saturday")))
         {
            return true;
         }
         else
         {
            return false;
         }
    }

    private int getNumericWeekday(String weekday)
    {
         if (weekday.equalsIgnoreCase("Sunday"))
             return 0;
         else if (weekday.equalsIgnoreCase("Monday"))
             return 1;
         else if (weekday.equalsIgnoreCase("Tuesday"))
             return 2;
         else if (weekday.equalsIgnoreCase("Wednesday"))
             return 3;
         else if (weekday.equalsIgnoreCase("Thursday"))
             return 4;
         else if (weekday.equalsIgnoreCase("Friday"))
             return 5;
         else if (weekday.equalsIgnoreCase("Saturday"))
             return 6;
         else
             return -1;
    }

    private String getStringWeekday(int weekday)
    {
         if (weekday == 0)
             return "Sunday";
         else if (weekday == 1)
             return "Monday";
         else if (weekday == 2)
             return "Tuesday";
         else if (weekday == 3)
             return "Wednesday";
         else if (weekday == 4)
             return "Thursday";
         else if (weekday == 5)
             return "Friday";
         else if (weekday == 6)
             return "Saturday";
         else
             return "";
    }

    public void run()
    {
        Calendar initialTimestamp = Calendar.getInstance();

        int lastMinute = initialTimestamp.get(Calendar.MINUTE);
        int lastHour = initialTimestamp.get(Calendar.HOUR_OF_DAY);
        int lastDay = initialTimestamp.get(Calendar.DAY_OF_MONTH);
          // MONTH is 0-based, so add 1
        int lastMonth = initialTimestamp.get(Calendar.MONTH) + 1;
        int lastWeekday = initialTimestamp.get(Calendar.DAY_OF_WEEK);

        int currentMinute;
        int currentHour;
        int currentDay;
        int currentMonth;
        int currentWeekday;

        boolean minuteChanged = false;
        boolean hourChanged = false;
        boolean dayChanged = false;
        boolean monthChanged = false;
        boolean weekdayChanged = false;

        STAFResult queueGetResult;

        for (;;)
        {
            queueGetResult = fHandle.submit2(
                "local", "QUEUE", "GET WAIT 10000");

            if (queueGetResult.rc == 0)
            {
                STAFQueueMessage queueMessage = new STAFQueueMessage(
                    queueGetResult.result);

                String queueType = queueMessage.type;

                if (queueType.equalsIgnoreCase(sQueueTypeEnd))
                {
                    // This type of queued message indicates that the service
                    // is terminating

                    try
                    {
                        fHandle.unRegister();
                    }
                    catch (STAFException ex)
                    {
                        if (DEBUG)
                        {
                            ex.printStackTrace();
                        }
                    }

                    return;
                }
                else if (queueType.equalsIgnoreCase("STAF/RequestComplete"))
                {
                    // A STAF/RequestComplete message is a map containing keys:
                        //   requestNumber, rc, result

                    Map messageMap = (Map)queueMessage.message;

                    String reqNum = (String)messageMap.get("requestNumber");
                    String rc = (String)messageMap.get("rc");
                    Object result = (Object)messageMap.get("result");
                    
                    // Log a message about the completed request

                    String level = "";

                    if (rc.equals("0"))
                        level = "info";
                    else
                        level = "error";

                    String request = (String)fSubmittedRequests.get(reqNum);

                    fHandle.submit2(
                        "local", "LOG", "LOG MACHINE LOGNAME " +
                        fServiceName + " LEVEL " + level + " MESSAGE " +
                        STAFUtil.wrapData("Request #" + reqNum +
                                          ": RC=" + rc +
                                          ", Result=" + result.toString() +
                                          ", Request submitted: " +request));

                    // handle process service start requests

                    String lowerRequest = request.toLowerCase();

                    if ((lowerRequest.indexOf("service=process") > -1) &&
                        (lowerRequest.indexOf("request=start") > -1)  &&
                        (lowerRequest.indexOf("wait") == -1))
                    {
                        String machine = queueMessage.machine;

                        String handle = result.toString();

                        STAFResult notifyRegResult;

                        notifyRegResult = fHandle.submit2(machine, "process",
                            "notify register onendofhandle " + handle);

                        fHandle.submit2(
                            "local", "log", "log machine logname " +
                            fServiceName + " level " + level +
                            " message " +
                            STAFUtil.wrapData("Register for process completion"
                                              + " notification RC=" +
                                              notifyRegResult.rc + " Result=" +
                                              notifyRegResult.result +
                                              " Request=" + machine + " process"
                                              +" notify register onendofhandle "
                                              + handle));

                        if (notifyRegResult.rc ==
                            STAFResult.ProcessAlreadyComplete)
                        {
                            STAFResult freeHandleResult;

                            freeHandleResult = fHandle.submit2(machine,
                                "process", "free handle " + handle);

                            fHandle.submit2(
                                "local", "log", "log machine logname "
                                + fServiceName + " level " + level +
                                " message " +
                                STAFUtil.wrapData("Free handle " + "RC=" +
                                                  freeHandleResult.rc +
                                                  " Result=" +
                                                  freeHandleResult.result +
                                                  " Request=" + machine +
                                                  " process free handle " +
                                                  handle));;
                        }
                    }

                    // Remove the request from the submitted request map

                    fSubmittedRequests.remove(reqNum);
                }
                else if (queueType.equalsIgnoreCase("STAF/PROCESS/END"))
                {
                    Map messageMap = (Map)queueMessage.message;
                    String handle = (String)messageMap.get("handle");

                    String machine = queueMessage.machine;

                    STAFResult freeHandleResult;

                    freeHandleResult = fHandle.submit2(machine, "process",
                        "free handle " + handle);

                    fHandle.submit2(
                        "local", "log", "log machine logname "
                        + fServiceName + " level info" + " message " +
                        STAFUtil.wrapData("Free handle " + "RC=" +
                                          freeHandleResult.rc + " Result=" +
                                          freeHandleResult.result + " Request="
                                          + machine + " process free handle " +
                                          handle));;
                }
            }

            minuteChanged = false;
            hourChanged = false;
            dayChanged = false;
            monthChanged = false;
            weekdayChanged = false;

            Calendar currentTimestamp = Calendar.getInstance();

            currentMinute = currentTimestamp.get(Calendar.MINUTE);
            currentHour = currentTimestamp.get(Calendar.HOUR_OF_DAY);
            currentDay = currentTimestamp.get(Calendar.DAY_OF_MONTH);
              // MONTH is 0-based, so add 1
            currentMonth = currentTimestamp.get(Calendar.MONTH) + 1;
            currentWeekday = currentTimestamp.get(Calendar.DAY_OF_WEEK);

            if (currentMinute != lastMinute)
            {
                minuteChanged = true;
                lastMinute = currentMinute;
            }
            else
            {
                minuteChanged = false;
            }

            if (currentHour != lastHour)
            {
                hourChanged = true;
                lastHour = currentHour;
            }
            else
            {
                hourChanged = false;
            }

            if (currentDay != lastDay)
            {
                dayChanged = true;
                lastDay = currentDay;
            }
            else
            {
                dayChanged = false;
            }

            if (currentMonth != lastMonth)
            {
                monthChanged = true;
                lastMonth = currentMonth;
            }
            else
            {
                monthChanged = false;
            }

            if (currentWeekday != lastWeekday)
            {
                weekdayChanged = true;
                lastWeekday = currentWeekday;
            }
            else
            {
                weekdayChanged = false;
            }

            if (minuteChanged || hourChanged || dayChanged ||
                monthChanged || weekdayChanged)
            {
                // create new Python Interpreter for each Cron
                fPythonInterpreter = new PythonInterpreter();

                CronData theData;

                for (Enumeration e = fRegTable.keys(); e.hasMoreElements();)
                {
                    Integer id = (Integer)e.nextElement();

                    theData = (CronData) fRegTable.get(id);

                    boolean minuteMatch = (theData.fMinute.contains(
                        String.valueOf(currentMinute)) ||
                        theData.fMinute.contains("ANY"));

                    boolean minuteMatchOrEmpty = theData.fMinute.isEmpty() ||
                        minuteMatch;

                    boolean hourMatch = (theData.fHour.contains(
                        String.valueOf(currentHour)) ||
                        theData.fHour.contains("ANY"));

                    boolean hourMatchOrEmpty = theData.fHour.isEmpty() ||
                        hourMatch;

                    boolean dayMatch = (theData.fDay.contains(
                        String.valueOf(currentDay)) ||
                        theData.fDay.contains("ANY"));

                    boolean dayMatchOrEmpty = theData.fDay.isEmpty() ||
                        dayMatch;

                    boolean monthMatch = (theData.fMonth.contains(
                        String.valueOf(currentMonth)) ||
                        theData.fMonth.contains("ANY"));

                    boolean monthMatchOrEmpty = theData.fMonth.isEmpty() ||
                        monthMatch;

                    boolean weekdayMatch = (theData.fWeekday.contains(
                        String.valueOf(currentWeekday)) ||
                        theData.fWeekday.contains("ANY"));

                    boolean weekdayMatchOrEmpty = theData.fWeekday.isEmpty() ||
                        weekdayMatch;

                    boolean minuteTrigger = minuteChanged && minuteMatch;

                    boolean hourTrigger = hourChanged && hourMatch;

                    boolean dayTrigger = dayChanged && dayMatch;

                    boolean monthTrigger = monthChanged && monthMatch;

                    boolean weekdayTrigger = weekdayChanged && weekdayMatch;

                    // All 5 time units must match or be empty, and there
                    // must be at least one trigger.
                    if ((minuteMatchOrEmpty && hourMatchOrEmpty &&
                         dayMatchOrEmpty && monthMatchOrEmpty &&
                         weekdayMatchOrEmpty) &&
                        (minuteTrigger || hourTrigger || dayTrigger ||
                         monthTrigger || weekdayTrigger))
                    {
                        pySetVar("STAFCronSubmit", "true");

                        Object submit = pyGetVar("STAFCronSubmit");

                        if ((submit != null) &&
                            (submit.toString().equals("true")))
                        {

                            String prepare = theData.fPrepare;
                            pyExec(prepare);

                            boolean pythonMachine = theData.fPythonMachine;
                            boolean pythonService = theData.fPythonService;
                            boolean pythonRequest = theData.fPythonRequest;

                            String machine = "";
                            String service = "";
                            String request = "";

                            if (pythonMachine)
                            {
                                machine = pyStringEval(theData.fMachine);
                            }
                            else
                            {
                                machine = theData.fMachine;
                            }

                            if (pythonService)
                            {
                                service = pyStringEval(theData.fService);
                            }
                            else
                            {
                                service = theData.fService;
                            }

                            if (pythonRequest)
                            {
                                request = pyStringEval(theData.fRequest);
                            }
                            else
                            {
                                request = theData.fRequest;
                            }

                            STAFResult matchResult = fHandle.submit2(
                                STAFHandle.ReqQueue, machine,
                                service, request);

                            fSubmittedRequests.put(
                                matchResult.result,
                                "machine=" + machine + " service=" + service +
                                " request=" + request);

                            String triggers = "";

                            if (minuteTrigger)
                            {
                                triggers += "minute[" + currentMinute + "],";
                            }

                            if (hourTrigger || hourMatch)
                            {
                                triggers += "hour[" + currentHour + "],";
                            }

                            if (dayTrigger || dayMatch)
                            {
                                triggers += "day[" + currentDay + "],";
                            }

                            if (monthTrigger || monthMatch)
                            {
                                triggers += "month[" + currentMonth + "],";
                            }

                            if (weekdayTrigger || weekdayMatch)
                            {
                                triggers += "weekday[" + currentWeekday + "],";
                            }

                            if (triggers.endsWith(","))
                            {
                                triggers = triggers.substring(0,
                                    triggers.length() - 1);
                            }

                            String message = "Submitted request #" +
                                matchResult.result + " ID=" + id.toString() +
                                " triggers=" + triggers +
                                " machine=" + machine + " service=" + service +
                                " request=" + request;

                            fHandle.submit2(
                                "local", "LOG", "LOG MACHINE LOGNAME " +
                                fServiceName + " LEVEL info MESSAGE " +
                                STAFUtil.wrapData(message));
                        }
                    }
                }
            }
        }

    }

    public void loadHashtable()
    {
        try
        {
            File fHashtableFile = new File(fHashtableFileName);
            if (!fHashtableFile.exists())
            {
                fRegTable = new Hashtable();
                return;
            }
            else
            {
                ObjectInputStream ois =
                    new ObjectInputStream(new FileInputStream(fHashtableFile));
                fRegTable = (Hashtable)(ois.readObject());
                ois.close();

                for (Enumeration e = fRegTable.elements();
                     e.hasMoreElements();)
                {
                    CronData data = (CronData) e.nextElement();

                    // XXX Do something here?
                }

                fCronID = findNextCronID();

            }
        }
        catch (IOException e)
        {
            fRegTable = new Hashtable();
            return;
        }
        catch (ClassNotFoundException e)
        {
            fRegTable = new Hashtable();
            return;
        }
    }

    public void saveHashtable()
    {
        try
        {
            ObjectOutputStream oos =
                new ObjectOutputStream(new
                    FileOutputStream(fHashtableFileName));
            oos.writeObject(((Object)(fRegTable)));
            oos.close();
        }
        catch (IOException ex)
        {
            ex.printStackTrace();
        }
    }

    public int findNextCronID()
    {
        int highestID = 0;
        for (Enumeration e = fRegTable.keys(); e.hasMoreElements();)
        {
            Integer id = (Integer) e.nextElement();
            if (id.intValue() > highestID)
            {
                highestID = id.intValue();
            }
        }

        return highestID + 1;
    }

    public STAFResult term()
    {
        fHandle.submit2(
            STAFHandle.ReqFireAndForget,
            "local", "QUEUE", "QUEUE TYPE " + sQueueTypeEnd +
            " MESSAGE " + STAFUtil.wrapData(""));

        return new STAFResult(STAFResult.Ok);
    }

    public String pyStringEval(String value)
    {
        synchronized(fPythonInterpreter)
        {
            PyObject result = null;
            try
            {
                result = fPythonInterpreter.eval(value);
            }
            catch (PyException ex)
            {
                ex.printStackTrace();
            }

            return result.toString();
        }
    }

    public void pySetVar(String varName, Object value)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                fPythonInterpreter.set(varName, value);
            }
            catch (PyException e)
            {
                fHandle.submit2(
                    "local", "LOG", "LOG MACHINE LOGNAME " +
                    fServiceName + " LEVEL error MESSAGE " +
                    STAFUtil.wrapData("PySetVar PyException caught setting " +
                    " var=" + varName + " to value=" + value + e.toString()));
            }
        }
    }

    public Object pyGetVar(String varName)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                return fPythonInterpreter.get(varName);
            }
            catch (PyException e)
            {
                fHandle.submit2(
                    "local", "LOG", "LOG MACHINE LOGNAME " +
                    fServiceName + " LEVEL error MESSAGE " +
                    STAFUtil.wrapData("PyGetVar PyException caught getting " +
                    " var=" + varName + " to value=" + varName +
                    e.toString()));

                return null;
            }
        }
    }

    public void pyExec(String value)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                fPythonInterpreter.exec(value);
            }
            catch (PyException e)
            {
                e.printStackTrace();
            }
        }
    }

}
