/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002, 2004                                        */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.stax;
import org.w3c.dom.Node;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import com.ibm.staf.*;
import com.ibm.staf.service.*;

public class STAXBlockActionFactory implements STAXActionFactory,
                                               STAXListRequestHandler,
                                               STAXQueryRequestHandler,
                                               STAXGenericRequestHandler,
                                               STAXJobManagementHandler
{
    static final String STAX_BLOCK_EVENT = new String("Block");

    static final String sBlockInfoMapClassName = new String(
        "STAF/Service/STAX/BlockInfo");
    static final String sQueryBlockMapClassName = new String(
        "STAF/Service/STAX/QueryBlock");

    // STAX Service Error Codes
    static final int BlockNotHeld = 4002;
    static final int BlockAlreadyHeld = 4003;

    private static String fDTDInfo =
"\n" +
"<!--================= The Block Element ============================ -->\n" +
"<!--\n" +
"     Defines a task block that can be held, released, or terminated.\n" +
"     Used in conjunction with the hold/terminate/release elements to\n" +
"     define a task block that can be held, terminated, or released.\n" +
"     The name attribute value is evaluated via Python.\n" +
"-->\n" +
"<!ELEMENT block      (%task;)>\n" +
"<!ATTLIST block\n" +
"          name       CDATA    #REQUIRED\n" +
">\n";

    public STAXBlockActionFactory(STAX staxService)
    {
        staxService.registerListHandler("BLOCKS", this);
        staxService.registerQueryHandler("BLOCK", "Block Name", this);
        staxService.registerQueryJobHandler(this);

        staxService.registerJobManagementHandler(this);

        // Set up Hold, Release, Terminate Block Parser

        fHRTParser.addOption("HOLD", 1, STAFCommandParser.VALUENOTALLOWED);
        fHRTParser.addOption("RELEASE", 1, STAFCommandParser.VALUENOTALLOWED);
        fHRTParser.addOption("TERMINATE", 1,STAFCommandParser.VALUENOTALLOWED);
        fHRTParser.addOption("JOB", 1, STAFCommandParser.VALUEREQUIRED);
        fHRTParser.addOption("BLOCK", 1, STAFCommandParser.VALUEREQUIRED);

        fHRTParser.addOptionGroup("HOLD RELEASE TERMINATE", 1, 1);
        fHRTParser.addOptionNeed("HOLD RELEASE TERMINATE", "JOB");

        // Register as a GenericRequestHandler
        try
        {
            // Assign STAFServiceInterfaceLevel class that this handler uses

            Class serviceInterfaceClass = Class.forName(STAX.INTERFACE_LEVEL_30);

            int rc = staxService.registerGenericRequestHandler(this,
                                                      serviceInterfaceClass);
            if (rc != 0)
            {
                System.out.println("STAXBlockActionFactory." +
                                   "registerGenericRequestHandler() failed");
            }
        }
        catch (ClassNotFoundException e)
        {
            System.out.println("STAXBlockActionFactory: " +
                               "registerGenericRequestHandler: " + e);
        }

        // Construct map-class for block information

        fBlockInfoMapClass = new STAFMapClassDefinition(sBlockInfoMapClassName);

        fBlockInfoMapClass.addKey("blockName", "Block Name");
        fBlockInfoMapClass.addKey("state",     "State");

        // Construct map-class for query block information

        fQueryBlockMapClass = new STAFMapClassDefinition(sQueryBlockMapClassName);

        fQueryBlockMapClass.addKey("blockName", "Block Name");
        fQueryBlockMapClass.addKey("state",     "State");
        fQueryBlockMapClass.addKey("threadID", "Thread ID");
        fQueryBlockMapClass.addKey("startTimestamp", "Start Date-Time");
    }

    public void initJob(STAXJob job)
    {
        if (!job.setData("blockMap", new TreeMap()))
        {
            String msg = "STAXBlockActionFactory.initJob: setData for " +
                         "blockMap failed.";
            job.log(STAXJob.JOB_LOG, "error", msg);
        }
    }

    public void terminateJob(STAXJob job)
    { /* Do Nothing */ }

    public STAFResult handleListRequest(String type, STAXJob job,
                                        STAXRequestSettings settings)
    {
        if (type.equalsIgnoreCase("blocks"))
        {
            // LIST JOB <Job ID> BLOCKS

            // Create the marshalling context and set its map class definitions
            // and create an empty list to contain the block map entries

            STAFMarshallingContext mc = new STAFMarshallingContext();
            mc.setMapClassDefinition(fBlockInfoMapClass);
            List blockOutputList = new ArrayList();

            // Iterate through the block map, generating the output list

            TreeMap blockMap = (TreeMap)job.getData("blockMap");

            synchronized (blockMap)
            {
                Iterator iter = blockMap.values().iterator();

                while (iter.hasNext())
                {
                    STAXBlockAction block = (STAXBlockAction)iter.next();

                    Map blockOutputMap = fBlockInfoMapClass.createInstance();
                    blockOutputMap.put("blockName", block.getName());
                    blockOutputMap.put("state", block.getBlockStateAsString());

                    blockOutputList.add(blockOutputMap);
                }
            }

            mc.setRootObject(blockOutputList);

            return new STAFResult(STAFResult.Ok, mc.marshall());
        }
        else
            return new STAFResult(STAFResult.DoesNotExist, type);
    }

    public STAFResult handleQueryRequest(String type, String key, STAXJob job,
                                        STAXRequestSettings settings)
    {
        if (type.equalsIgnoreCase("block"))
        {
            // QUERY JOB <Job ID> BLOCK <Block name>

            // Create the marshalling context and set its map class definitions

            STAFMarshallingContext mc = new STAFMarshallingContext();
            mc.setMapClassDefinition(fQueryBlockMapClass);

            TreeMap blockMap = (TreeMap)job.getData("blockMap");

            synchronized (blockMap)
            {
                if (!blockMap.containsKey(key))
                    return new STAFResult(STAFResult.DoesNotExist, key);
                else
                {
                    STAXBlockAction block = (STAXBlockAction)blockMap.get(key);

                    Map blockOutputMap = fQueryBlockMapClass.createInstance();
                    blockOutputMap.put("blockName", block.getName());
                    blockOutputMap.put("state", block.getBlockStateAsString());
                    blockOutputMap.put("threadID", "" + block.getOwningThread().
                                       getThreadNumber());
                    blockOutputMap.put("startTimestamp",
                                       block.getStartTimestamp().
                                       getTimestampString());

                    mc.setRootObject(blockOutputMap);

                    return new STAFResult(STAFResult.Ok, mc.marshall());
                }
            }
        }
        else
            return new STAFResult(STAFResult.DoesNotExist, type);
    }

    public STAFResult handleQueryJobRequest(STAXJob job,
                                            STAXRequestSettings settings)
    {
        // Provide additional information for the QUERY JOB <Job ID> result

        int blocksRunningState = 0;
        int blocksHeldState = 0;
        int blocksUnknownState = 0;

        TreeMap blockMap = (TreeMap)job.getData("blockMap");
        synchronized (blockMap)
        {
            Iterator iter = blockMap.values().iterator();

            while (iter.hasNext())
            {
                STAXBlockAction block = (STAXBlockAction)iter.next();
                if (block.getBlockState() == STAXBlockAction.BLOCK_RUNNING)
                    blocksRunningState += 1;
                else if (block.getBlockState() ==
                                             STAXBlockAction.BLOCK_HELD)
                    blocksHeldState += 1;
                else
                    blocksUnknownState += 1;
            }
        }

        // XXX: Change to pass in the marshalling context 
        STAFMarshallingContext mc = new STAFMarshallingContext();

        Map blockOutputMap = new HashMap();
        blockOutputMap.put("numBlocksRunning", "" + blocksRunningState);
        blockOutputMap.put("numBlocksHeld", "" + blocksHeldState);
        blockOutputMap.put("numBlocksUnknown", "" + blocksUnknownState);

        mc.setRootObject(blockOutputMap);

        return new STAFResult(STAFResult.Ok, mc.marshall()); 
    }

    public String getDTDInfo()
    {
        return fDTDInfo;
    }

    public String getDTDTaskName()
    {
        return "block";
    }

    public STAXAction parseAction(STAX staxService, STAXJob job,
                                  org.w3c.dom.Node root) throws STAXException
    {
        String blockName = new String();
        STAXAction blockAction = null;

        NamedNodeMap attrs = root.getAttributes();

        for (int i = 0; i < attrs.getLength(); ++i)
        {
            Node thisAttr = attrs.item(i);

            String errorInfo = "\n  Element: " + root.getNodeName() +
                               "  Attribute: " + thisAttr.getNodeName();

            if (thisAttr.getNodeName().equals("name"))
            {
                blockName = STAXUtil.parseAndCompileForPython(
                    thisAttr.getNodeValue(), errorInfo);
            }
            else
            {
                throw new STAXInvalidXMLAttributeException(
                          root.getNodeName() + ": " + thisAttr.getNodeName());
            }
        }

        NodeList children = root.getChildNodes();

        for (int i = 0; i < children.getLength(); ++i)
        {
            Node thisChild = children.item(i);

            if (thisChild.getNodeType() == Node.COMMENT_NODE)
            {
                /* Do nothing */
            }
            else if (thisChild.getNodeType() == Node.ELEMENT_NODE)
            {
                if (blockAction != null)
                {
                    throw new STAXInvalidXMLElementCountException(
                              thisChild.getNodeName());
                }

                STAXActionFactory factory =
                    staxService.getActionFactory(thisChild.getNodeName());

                if (factory == null)
                {
                    throw new STAXInvalidXMLElementException(
                              thisChild.getNodeName());
                }

                blockAction = factory.parseAction(staxService, job, thisChild);
            }
            else
            {
                throw new STAXInvalidXMLNodeTypeException(
                          Integer.toString(thisChild.getNodeType()));
            }
        }

        return new STAXBlockAction(blockName, blockAction);
    }


    // STAXGenericRequestHandler Interface Methods

    public STAFResult handleRequest(Object infoObject, STAX staxService)
    {
        STAFServiceInterfaceLevel30.RequestInfo info =
            (STAFServiceInterfaceLevel30.RequestInfo)infoObject;

        // Parse the request

        STAFCommandParseResult parseResult= fHRTParser.parse(info.request);

        if (parseResult.rc != STAFResult.Ok)
        {
            if ((parseResult.numInstances() >= 1) &&
                ((parseResult.instanceName(1).equalsIgnoreCase("HOLD")) ||
                 (parseResult.instanceName(1).equalsIgnoreCase("RELEASE")) ||
                 (parseResult.instanceName(1).equalsIgnoreCase("TERMINATE"))))
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      parseResult.errorBuffer);
            }
            else
            {   // Returning nothing in the result indicates that this parser
                // does not support this request.
                return new STAFResult(STAFResult.InvalidRequestString, "");
            }
        }

        // Verify the requesting machine/user has at least trust level 4

        String request = "";

        if (parseResult.optionTimes("HOLD") != 0)
            request = "HOLD";
        else if (parseResult.optionTimes("RELEASE") != 0)
            request = "RELEASE";
        else if (parseResult.optionTimes("TERMINATE") != 0)
            request = "TERMINATE";

        STAFResult trustResult = STAFUtil.validateTrust(
            4, staxService.getServiceName(), request,
            staxService.getLocalMachineName(), info);

        if (trustResult.rc != STAFResult.Ok) return trustResult;

        STAXJob job = null;

        try
        {
            job = (STAXJob)staxService.getJobMap().get(
                  new Integer(parseResult.optionValue("JOB")));

            if (job == null)
            {
                return new STAFResult(STAFResult.DoesNotExist,
                                      parseResult.optionValue("JOB"));
            }
        }
        catch (NumberFormatException e)
        {
            return new STAFResult(STAFResult.InvalidValue,
                                  parseResult.optionValue("JOB"));
        }

        String blockName = new String("main");

        if (parseResult.optionTimes("BLOCK") != 0)
            blockName = parseResult.optionValue("BLOCK");

        TreeMap blockMap = (TreeMap)job.getData("blockMap");

        if (!blockMap.containsKey(blockName))
            return new STAFResult(STAFResult.DoesNotExist, blockName);

        STAXBlockAction theBlock = (STAXBlockAction)blockMap.get(blockName);

        if (parseResult.optionTimes("HOLD") != 0)
        {
            if (theBlock.getBlockState() == STAXBlockAction.BLOCK_HELD)
                return new STAFResult(BlockAlreadyHeld,
                                      "Block already held");

            String msg = "Received HOLD BLOCK "  + blockName + " request";
            job.log(STAXJob.JOB_LOG, "info", msg);

            theBlock.holdBlock();
        }
        else if (parseResult.optionTimes("RELEASE") != 0)
        {
            if (theBlock.getBlockState() != STAXBlockAction.BLOCK_HELD)
                return new STAFResult(BlockNotHeld, "Block not held");

            String msg = "Received RELEASE BLOCK " + blockName + " request";
            job.log(STAXJob.JOB_LOG, "info", msg);

            theBlock.releaseBlock();
        }
        else
        {
            String msg = "Received TERMINATE BLOCK "  + blockName + " request";
            job.log(STAXJob.JOB_LOG, "info", msg);

            theBlock.terminateBlock();
        }

        return new STAFResult(STAFResult.Ok);
    }

    public String getHelpInfo(String lineSep)
    {
        return "HOLD      JOB <Job ID> [BLOCK <Block Name>]" +
               lineSep + lineSep +
               "RELEASE   JOB <Job ID> [BLOCK <Block Name>]" +
               lineSep + lineSep +
               "TERMINATE JOB <Job ID> [BLOCK <Block Name>]";
    }

    private STAFCommandParser fHRTParser = new STAFCommandParser();
    private STAFMapClassDefinition fBlockInfoMapClass;
    private STAFMapClassDefinition fQueryBlockMapClass;
}
