/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2004, 2005                                        */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.stax.staxdoc;

import java.util.*;
import java.io.*;

/******************************************************************************
* Main program of StaxDoc tool.
******************************************************************************/
public final class Main
{
  static final String program = "staxdoc";

  static Options options = new Options();
  static StaxSource srcFiles = null;

  /////////////////////////////////////////////////////////////////////////////
  public static void main(String argv[])
  {
    //-------------------------------------------------------------------------
    // Parse arguments

    // first looks for 'verbose' option
    for (int i = 0; i < argv.length; i++)
    {
      String arg = argv[i];
      if (argv[i].equals("-verbose"))
      {
        options.verbose = true;
        System.out.println("Option set: verbose=true");
      }
    }

    for (int i = 0; i < argv.length; i++)
    {
      String arg = argv[i];
      if (arg.equals("-d"))
      {
        i++;
        options.destDir = argv[i];
        if (options.verbose) System.out.println("Option set: d=" + options.destDir);

        File destDir = new File(argv[i]);
        if (!destDir.exists())
        {
          //System.out.println("Destination directory not found " + destDir.getPath());
          if (!destDir.mkdirs())
          {
            System.err.println("Unable to create " + destDir.getName() + " directory");
            exit();
          }
          if (options.verbose) System.out.println("Destination directory created: " + destDir.getName());
        }
        else if (!destDir.isDirectory())
        {
          System.err.println("Destination directory is not a directory" + destDir.getPath());
          exit();
        }
        else if (!destDir.canWrite())
        {
          System.err.println("Destination directory not writable" + destDir.getPath());
          exit();
        }
      }
      else if (arg.equals("-doctitle"))
      {
        i++;
        options.doctitle = argv[i];
        if (options.verbose) System.out.println("Option set: doctitle=" + options.doctitle);
      }
      else if (arg.equals("-help"))
      {
        printUsage();
        exit();
      }
      else if (arg.equals("-overview"))
      {
        i++;
        options.overviewFilename = argv[i];
        if (options.verbose) System.out.println("Option set: overview=" + options.overviewFilename);
      }
      else if (arg.equals("-sourcepath"))
      {
        i++;
        options.srcPathString = argv[i];
        if (options.verbose) System.out.println("Option set: sourcepath=" + options.srcPathString);
      }
      else if (arg.equals("-verbose"))
      {
        // nothing to do
      }
      else if (arg.equals("-windowtitle"))
      {
        i++;
        options.windowtitle = argv[i];
        if (options.verbose) System.out.println("Option set: windowtitle=" + options.windowtitle);
      }
      else if ( arg.startsWith("-") )
      {
        usageError("Invalid option: " + arg);
      }
      else
      {
        StringTokenizer st = new StringTokenizer(arg, "=");
        String pkgPath = st.nextToken();
        String pkgName = pkgPath;
        if (st.hasMoreTokens())
        {
          pkgName = st.nextToken();
        }

        options.packagePathsList.add(pkgPath);
        options.packageNamesList.add(pkgName);
        if (options.verbose) System.out.println("STAX source package:" + arg);
      }
    }

    if (options.packagePathsList.isEmpty())
    {
      usageError("No packages specified");
    }

    //-------------------------------------------------------------------------
    // put all STAX source files into global StaxFiles object
    srcFiles = new StaxSource(options.srcPathString);
    try
    {
      System.out.println();
      for(int i=0; i<options.packagePathsList.size(); i++)
      {
        srcFiles.addSource((String)options.packagePathsList.get(i), (String)options.packageNamesList.get(i));
      }
      System.out.println();
    }
    catch (IOException e)
    {
      exit();
    }

    //-------------------------------------------------------------------------
    // generate documentation files
    try
    {
      WriterMainFrame.generate();
      WriterOverviewSummary.generate();
      WriterOverviewFrame.generate();
      WriterAllFilesFrame.generate();

      // generate packages directories structure and navigation HTML files
      Iterator itDir = Main.srcFiles.getPackageIterator();
      while (itDir.hasNext())
      {
        new PackageHandler((StaxPackage)itDir.next());
      }

      // generate files
      Iterator itFile = Main.srcFiles.getFilesIterator();

      while (itFile.hasNext())
      {
        WriterStaxDoc.generate((StaxFile)itFile.next());
      }

    }
    catch(IOException ioe)
    {
      System.err.println("IOException encountered");
      System.err.println(ioe.toString());
      exit();
    }


    if (options.verbose) System.out.println("STAXDoc ended with success");

    exit();
  }


  /////////////////////////////////////////////////////////////////////////////
  static void exit()
  {
    System.exit(0);
  }


  /////////////////////////////////////////////////////////////////////////////
  static void printUsage()
  {
    System.out.println();
    System.out.println("Usage: java -jar STAXDoc.jar [-options] packagenames...");
    System.out.println();
    System.out.println("where options include:");
    System.out.println("    -d <directory>          Destination directory for output files");
    System.out.println("                            The current directory is the default");
    System.out.println("    -doctitle <html-code>   Include title for the package index (first) page");
    System.out.println("    -help                   Display command line options");
    System.out.println("    -overview <file>        Read overview documentation from HTML file");
    System.out.println("    -sourcepath <directory> Root directory of the packages");
    System.out.println("                            The current directory is the default");
    System.out.println("    -verbose                Output messages about what STAXDoc is doing");
    System.out.println("    -windowtitle <title>    The title to be placed in the HTML <title> tag.");
    System.out.println();
    System.out.println("where packagenames specifies the names of one or more subdirectories in");
    System.out.println("    the -sourcepath containing STAX xml files that you want to document");
    System.out.println();
    System.out.println("Example:");
    System.out.println("  java -jar STAXDoc.jar -d C:\\stax\\mydocs -sourcepath C:\\stax\\xml src1 src2");
    System.out.println();
  }
  

  /////////////////////////////////////////////////////////////////////////////
  static void usageError(String key)
  {
    System.err.println(key);
    printUsage();
    exit();
  }

}
