/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFString.h"
#include "STAFUtil.h"
#include "STAFUtilWin32.h"
#include "STAFMutexSem.h"
#include <time.h>

char *STAFUtilGetCurrentProcessCodePage(char *codepage)
{
    if (codepage != 0)
        sprintf(codepage, "%s%d", "IBM-", GetOEMCP());

    return codepage;
}


unsigned int STAFUtilWin32GetWinType()
{
    static STAFMutexSem helperSem;
    static bool haveChecked = false;
    static unsigned int type;

    if (!haveChecked)
    {
        STAFMutexSemLock lock(helperSem);

        if (!haveChecked)
        {
            OSVERSIONINFO osVersionInfo = { sizeof(OSVERSIONINFO), 0 };

            if (!GetVersionEx(&osVersionInfo))
            {
                cout << "GetVersionEx() failed in STAFUtilWin32GetWinType()"
                     << endl;
                type = 0;
            }
            else if (osVersionInfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
            {
                if (osVersionInfo.dwMinorVersion == 0)
                    type = kSTAFWin95;
                else if (osVersionInfo.dwMinorVersion == 10)
                    type = kSTAFWin98;
                else if (osVersionInfo.dwMinorVersion == 90)
                    type = kSTAFWinMe;
                else
                    type = kSTAFWin9XUnknown;
            }
            else if (osVersionInfo.dwPlatformId == VER_PLATFORM_WIN32_NT)
            {
                if (osVersionInfo.dwMajorVersion == 4)
                    type = kSTAFWinNT;
                else if ((osVersionInfo.dwMajorVersion == 5) &&
                         (osVersionInfo.dwMinorVersion == 0))
                    type = kSTAFWin2K;
                else if ((osVersionInfo.dwMajorVersion == 5) &&
                         (osVersionInfo.dwMinorVersion == 1))
                    type = kSTAFWinXP;
                else if ((osVersionInfo.dwMajorVersion == 5) &&
                         (osVersionInfo.dwMinorVersion == 2))
                    type = kSTAFWin2K3;
                else
                    type = kSTAFWinNTUnknown;
            }
            else
            {
                type = kSTAFWin32Unknown;
            }

            haveChecked = true;
        }
    }

    return type;
}


unsigned int STAFUtilGetConfigInfo(STAFConfigInfo *configInfo,
                                   STAFString_t *errorBuffer,
                                   unsigned int *osRC)
{
    // Get total physical memory

    MEMORYSTATUS memoryStatus = { sizeof(MEMORYSTATUS), 0 };

    GlobalMemoryStatus(&memoryStatus);

    configInfo->physicalMemory = memoryStatus.dwTotalPhys;

    // Get boot drive

    char systemPathBuffer[MAX_PATH];

    if (GetSystemDirectory(systemPathBuffer, MAX_PATH) == 0)
    {
        if (errorBuffer)
            *errorBuffer = STAFString("GetSystemDirectory()").adoptImpl();

        if (osRC) *osRC = static_cast<unsigned int>(GetLastError());

        return kSTAFBaseOSError;
    }

    STAFString bootDrive = STAFString(systemPathBuffer).subString(0, 2);

    // Get STAFRoot

    char buffer[256] = { 0 };
    char *filePart = 0;

    if (SearchPath(0, "STAFPROC.EXE", 0, sizeof(buffer), buffer,
        &filePart) == 0)
    {
        if (errorBuffer)
            *errorBuffer = STAFString("SearchPath()").adoptImpl();

        if (osRC) *osRC = static_cast<unsigned int>(GetLastError());

        return kSTAFBaseOSError;
    }

    STAFString exePath = STAFString(buffer).subWord(0);

    exePath = exePath.subString(0, exePath.findLastOf(kUTF8_BSLASH));
    exePath = exePath.subString(0, exePath.findLastOf(kUTF8_BSLASH));

    // Get OS Name, Major and Minor Version, and Revision

    OSVERSIONINFO osVersionInfo = { sizeof(OSVERSIONINFO), 0 };

    if (GetVersionEx(&osVersionInfo) != TRUE)
    {
        if (errorBuffer)
            *errorBuffer = STAFString("GetVersionEx()").adoptImpl();

        if (osRC) *osRC = 1;

        return kSTAFBaseOSError;
    }

    STAFString osMajorVersion = STAFString(osVersionInfo.dwMajorVersion);
    STAFString osMinorVersion = STAFString(osVersionInfo.dwMinorVersion);
    STAFString osRevision = STAFString(osVersionInfo.dwBuildNumber &
                                       0x0000FFFF);
    STAFString osName;

    switch (STAFUtilWin32GetWinType())
    {
        case kSTAFWin95:        osName = "Win95"; break;
        case kSTAFWin98:        osName = "Win98"; break;
        case kSTAFWinMe:        osName = "WinMe"; break;
        case kSTAFWin9XUnknown: osName = "Unknown Win9X"; break;
        case kSTAFWinNT:        osName = "WinNT"; break;
        case kSTAFWin2K:        osName = "Win2000"; break;
        case kSTAFWinXP:        osName = "WinXP"; break;
        case kSTAFWin2K3:       osName = "Win2003"; break;
        case kSTAFWinNTUnknown: osName = "Unknown WinNT"; break;
        case kSTAFWin32Unknown: osName = "Unknown Win32"; break;
    }

    STAFString crlfString = STAFString(kUTF8_CR) + STAFString(kUTF8_LF);

    configInfo->lineSeparator  = crlfString.adoptImpl();
    configInfo->fileSeparator  = STAFString(kUTF8_BSLASH).adoptImpl();
    configInfo->pathSeparator  = STAFString(kUTF8_SCOLON).adoptImpl();
    configInfo->commandSeparator  = STAFString(kUTF8_AMP).adoptImpl();
    configInfo->osName         = osName.adoptImpl();
    configInfo->osMajorVersion = osMajorVersion.adoptImpl();
    configInfo->osMinorVersion = osMinorVersion.adoptImpl();
    configInfo->osRevision     = osRevision.adoptImpl();
    configInfo->exePath        = exePath.adoptImpl();
    configInfo->bootDrive      = bootDrive.adoptImpl();
    configInfo->defaultProcessStopMethod = kSTAFProcessStopWithSigKill;
    configInfo->envVarCaseSensitive = 0;  // Case insensitive

    return 0;
}


void *STAFUtilGetSystemMemory(unsigned long size, unsigned int *osRC)
{
    void *theMem = GlobalAlloc(GMEM_FIXED, size);

    if ((theMem == 0) && osRC) *osRC = 0;

    return theMem;
}


void STAFUtilFreeSystemMemory(void *ptr)
{
    GlobalFree(ptr);
}


unsigned int STAFUtilIsValidSocket(STAFSocket_t theSocket)
{
    return (theSocket != INVALID_SOCKET);
}


unsigned int STAFUtilGetNonInheritableSocket(STAFSocket_t oldSocket,
                                             STAFSocket_t *newSocket,
                                             unsigned int *osRC)
{
    HANDLE newHandle;

    int dupHandleRC = DuplicateHandle(
        GetCurrentProcess(), (HANDLE)oldSocket, 
        GetCurrentProcess(), &newHandle, 0,
        FALSE, // The new handle cannot be inherited
        DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS);

    if (dupHandleRC)
    {
        *newSocket = (SOCKET)newHandle;
        return kSTAFOk;  // To indicate success
    }
    else
    {
        // DuplicateHandle failed
        *osRC = static_cast<unsigned int>(GetLastError());
        *newSocket = oldSocket;
        return 1;        // To indicate failure
    }
}


unsigned int STAFUtilWin32CreateNullSD(PSECURITY_DESCRIPTOR *sd)
{
    // If on WinNT or above, set up security to use on a file mapping to
    // allow access to everyone.

    if (STAFUtilWin32GetWinType() & kSTAFWinNTPlus)
    {
        // Allocate the Security Descriptor (SD)
        if (!(*sd = static_cast<PSECURITY_DESCRIPTOR>(
                    new char[SECURITY_DESCRIPTOR_MIN_LENGTH])))
            return 1;

        if (!InitializeSecurityDescriptor(*sd, SECURITY_DESCRIPTOR_REVISION))
            return 2;

        // Set NULL DACL on the Security Descriptor
        if (!SetSecurityDescriptorDacl(*sd, TRUE, (PACL)NULL, FALSE))
            return 3;
    }

    return 0;
}


unsigned int STAFUtilWin32DeleteNullSD(PSECURITY_DESCRIPTOR *sd)
{
    // If on WinNT or above, delete the Security Descriptor

    if (STAFUtilWin32GetWinType() & kSTAFWinNTPlus)
    {
        if (*sd != 0)
        {
            delete [] static_cast<char *>(*sd);
            *sd = 0;
        }
    }

    return 0;
}


// The Win32 API does not contain a functional equivalent of the standard C
// mkstemp() function.  The Microsoft C runtime implementation does not even
// provide support for the function, although it does provide an
// implementation of mktemp().  However, it is strongly advised not to use
// the mktemp() function on either Unix or Windows.
//
// The Win32 API does provide a function, GetTempFileName(), that will
// generate a temporary filename, but that is all that it does; it does not
// open the file for you.  So, we are not using this function.
// Instead of using GetTempPath() to obtain the current user's setting for the
// location to place temporary files, the location to store temporary files is 
// being passed in (and should be *gSTAFWriteLocationPtr, the STAF data
// directory).
// 
// The STAFUtilCreateTempFile() method generates our own random filename, 
// using the system time to seed the rand() function.  The format of the
// unique filename is:  
//   <tempDir>\STAFTemp??????[.<suffix>]
// where <tempDir> contains the name of the directory to store the temp file
//       ??????    represents 6 randomly generated characters.
// Also, if a suffix is specified, e.g. "out", it is added as the file
// extension.
// We attempt to create the file.  If not successful, we repeat the process,
// generating another random temporary filename and attempting to create the
// file until the file is successfully created or some kind of fatal error
// occurs.

STAFRC_t STAFUtilCreateTempFile(STAFStringConst_t tempDir,
                                STAFStringConst_t tempFileSuffix,
                                STAFString_t *tempFileName,
                                STAFString_t *errorBuffer,
                                unsigned int *osRC)
{
    static char fileNameChars[37] = "0123456789ABCDEFGHIJKLMNOPQRSTUVYXWZ";
    const unsigned int BUFSIZE = 4096;    // Maximum buffer size
    char tmpPathBuffer[BUFSIZE] = { 0 };  // Buffer for temp path name
    char tmpFileBuffer[MAX_PATH] = { 0 }; // Buffer for temp file name
    char randomBuffer[7] = { 0 };         // Buffer for random values
    unsigned int characterRange = lstrlen(fileNameChars) - 1;
    char cCharacter;
    unsigned int i;
    HANDLE hFile;    // Temporary file's handle

    // Create a unique temporary file name with format:
    //   <tempDir>\STAFTemp??????[.<suffix>]
    // where <tempDir> contains the name of the directory to store the
    // temporary file and ?????? is a randomly assigned unique value.
    // Also, if a suffix is specified, e.g. out, it is added as the file
    // extension.

    STAFString tempNamePrefix = STAFString(tempDir) + STAFString(kUTF8_BSLASH) +
        "STAFTemp";
    
    STAFString suffix = STAFString(tempFileSuffix);

    srand((unsigned)time(NULL));

    // Repeat the process of generating a random temporary file name and
    // attempting to create the file, until the file is successfully created
    // or some kind of fatal error occurs.
    do
    {
        // Add 6 randomly assigned values to the temporary file name

        for (i = 0; i < 6; i++)
        {
            cCharacter = fileNameChars[rand() % characterRange];
            randomBuffer[i] = cCharacter;
        }

        randomBuffer[i] = 0;
        
        STAFString tempNameString = tempNamePrefix + STAFString(randomBuffer);
        
        // If a suffix was specified, add it to the temporary file name

        if (suffix.length() > 0)
        {
            tempNameString = tempNameString + STAFString(".") + suffix;
        }

        // Attempt to create the file to make sure doesn't already exist

        strncpy(tmpPathBuffer, tempNameString.toCurrentCodePage()->buffer(),
                sizeof(tmpPathBuffer));
        
        hFile = CreateFile(tempNameString.toCurrentCodePage()->buffer(),
                           GENERIC_READ | GENERIC_WRITE,
                           FILE_SHARE_READ | FILE_SHARE_WRITE,
                           0, CREATE_NEW, FILE_ATTRIBUTE_TEMPORARY, 0);

        if ((hFile == INVALID_HANDLE_VALUE) &&
            (GetLastError() != ERROR_FILE_EXISTS))
        {
            // Fatal error - Don't retry

            *errorBuffer = STAFString("Creating temp file failed").adoptImpl();
            *osRC = GetLastError();

            return kSTAFBaseOSError;
        }
    } while (hFile == INVALID_HANDLE_VALUE);

    // Temporary file was successfully created and opened.  Close it.
    CloseHandle(hFile);

    *tempFileName = STAFString(tmpPathBuffer).adoptImpl();

    return kSTAFOk;
}
