/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFString.h"
#include "STAFProc.h"
#include "STAFException.h"
#include "STAFConnectionManager.h"
#include "STAFConnectionProvider.h"
#include "STAFThreadManager.h"
#include "STAFTrace.h"

STAFConnectionManager::STAFConnectionManager()
{
    /* Do Nothing */
}


STAFRC_t STAFConnectionManager::addConnectionProvider(
    const STAFString &name,
    const STAFString &library,
    const ConnectionProviderOptionList &optionList,
    STAFString &errorBuffer)
{
    STAFConnectionProviderPtr theConnProv;

    try
    {
        // Check if connection provider with this name already exists

        if (fConnProvMap.find(name.toLowerCase()) != fConnProvMap.end())
        {
            errorBuffer = "Interface " + name.toLowerCase() + " already exists";
            return kSTAFAlreadyExists;
        }

        STAFConnectionProviderConstructInfoLevel1 constructInfo =
        { kSTAFConnectionProviderDuplex, 0 };

        STAFBuffer<STAFString> optionNames(new STAFString[optionList.size()],
                                           STAFBuffer<STAFString>::INIT,
                                           STAFBuffer<STAFString>::ARRAY);
        STAFBuffer<STAFString> optionValues(new STAFString[optionList.size()],
                                            STAFBuffer<STAFString>::INIT,
                                            STAFBuffer<STAFString>::ARRAY);

        STAFBuffer<STAFStringConst_t> optionNameImpls(
            new STAFStringConst_t[optionList.size()],
            STAFBuffer<STAFStringConst_t>::INIT,
            STAFBuffer<STAFStringConst_t>::ARRAY);
        STAFBuffer<STAFStringConst_t> optionValueImpls(
            new STAFStringConst_t[optionList.size()],
            STAFBuffer<STAFStringConst_t>::INIT,
            STAFBuffer<STAFStringConst_t>::ARRAY);

        for (int i = 0; i < optionList.size(); ++i)
        {
            unsigned int equalPos = optionList[i].find(kUTF8_EQUAL);

            optionNames[i] = optionList[i].subString(0, equalPos);
            optionValues[i] = (equalPos == STAFString::kNPos) ?
                                  STAFString() :
                                  optionList[i].subString(equalPos + 1);

            optionNameImpls[i] = optionNames[i].getImpl();
            optionValueImpls[i] = optionValues[i].getImpl();
        }

        constructInfo.numOptions = optionList.size();
        constructInfo.optionNames = optionNameImpls;
        constructInfo.optionValues = optionValueImpls;

        STAFConnectionProviderPtr theConnProv = 
            STAFConnectionProvider::createRefPtr(name, library,
                                                 &constructInfo, 1);
        // XXX: Hmmm... Not sure we should be doing this here
        // theConnProv->start(HandleRequest);

        fConnProvMap[name.toLowerCase()] = theConnProv;
        fConnProvList.push_back(theConnProv);

        if ((fDefaultConnProv == "") || (fDefaultConnProv == "local"))
            fDefaultConnProv = name;
    }
    catch (STAFException &se)
    {
        errorBuffer = se.getText();
        return se.getErrorCode();
    }

    return kSTAFOk;
}


STAFRC_t STAFConnectionManager::setDefaultConnectionProvider(
    const STAFString &name)
{
    ConnectionProviderMap::iterator iter = fConnProvMap.find(name.toLowerCase());

    if (iter == fConnProvMap.end())
        return kSTAFDoesNotExist;
    
    fDefaultConnProv = name;

    return kSTAFOk;
}


STAFString STAFConnectionManager::getDefaultConnectionProvider()
{
    return fDefaultConnProv;
}


STAFConnectionManager::ConnectionProviderList
    STAFConnectionManager::getConnectionProviderListCopy()
{
    return fConnProvList;
}


STAFRC_t STAFConnectionManager::makeConnection(const STAFString &where,
                                               STAFConnectionPtr &connection,
                                               STAFString &errorBuffer)
{
    STAFConnectionProviderPtr notUsed;

    return makeConnection(where, notUsed, connection, errorBuffer);
}


STAFRC_t STAFConnectionManager::makeConnection(
    const STAFString &where, STAFConnectionProviderPtr &provider,
    STAFConnectionPtr &connection, STAFString &errorBuffer)
{
    static STAFString sLocal("local");
    STAFRC_t rc = kSTAFOk;
    STAFString endpoint = where;
    STAFString connProvName = fDefaultConnProv;
    unsigned int sepPos = where.find(gSpecSeparator);

    if (sepPos != STAFString::kNPos)
    {
        connProvName = where.subString(0, sepPos);
        endpoint = where.subString(sepPos + gSpecSeparator.length());
    }
    else if (where.toLowerCase() == sLocal)
    {
        connProvName = sLocal;
    }

    // Attempt to make a connection up to the number of times specified by
    // gConnectionAttempts until a connection is made.

    for (int attempt = 1; attempt <= gConnectionAttempts; ++attempt)
    {
        rc = kSTAFOk;

        try
        {
            ConnectionProviderMap::iterator iter =
                fConnProvMap.find(connProvName.toLowerCase());

            if (iter == fConnProvMap.end())
            {
                errorBuffer = "No such interface: " + connProvName +
                    ", Endpoint: " + where;
                return kSTAFNoPathToMachine;
            }

            provider = iter->second;
            connection = iter->second->connect(endpoint);
        }
        catch (STAFException &e)
        {
            rc = kSTAFNoPathToMachine;
            errorBuffer = e.getText() + STAFString(": ") +
                STAFString(e.getErrorCode()) + ", Endpoint: " + where;
        }
        catch (...)
        {
            rc = kSTAFUnknownError;
        }

        if (rc == kSTAFOk) return rc;  // Worked, so don't retry

        int delayTime = 0;

        // Sleep (except after last connection attempt)
        if (attempt < gConnectionAttempts)
        {
            // Sleep a random number of milliseconds between 1 and the
            // gConnectionRetryDelay to reduce the # of connection attempts
            // needed when getting RC 16 recv: 111 errors that occur when the
            // queue for pending socket requests is full (Bug #915342)

            delayTime = rand() % gConnectionRetryDelay;
            STAFThreadManager::sleepCurrentThread(delayTime);
        }

        STAFTrace::trace(
            kSTAFTraceWarning,
            "STAFConnectionManager::makeConnection - Attempt #" +
            STAFString(attempt) + " of " + gConnectionAttempts +
            " (Delay " + delayTime + " milliseconds), RC: " + rc +
            ", Result: " + errorBuffer);
    }

    return rc;
}


STAFConnectionManager::~STAFConnectionManager()
{
    /* Do Nothing */
}
