/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001, 2004, 2005                                  */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/
 
#include "STAF.h"
#include "STAFProc.h"
#include "STAFHandleService.h"
#include "STAFHandleManager.h"
#include "STAFProcUtil.h"
#include "STAFServiceManager.h"
#include "STAFTrace.h"

STAFHandleService *STAFHandleService::sHandleService = 0;

STAFHandleService::STAFHandleService() : STAFService("HANDLE")
{
    sHandleService = this;

    // query options

    fQueryParser.addOption("QUERY",      1,
        STAFCommandParser::kValueNotAllowed);
    fQueryParser.addOption("HANDLE",    1,
        STAFCommandParser::kValueRequired);
 
    // query groups

    fQueryParser.addOptionGroup("QUERY",    1, 1);

    // query needs

    fQueryParser.addOptionNeed("QUERY", "HANDLE");

    // create options

    fCreateParser.addOption("CREATE",      1,
        STAFCommandParser::kValueNotAllowed);
    fCreateParser.addOption("HANDLE",      1,
        STAFCommandParser::kValueNotAllowed);
    fCreateParser.addOption("NAME",        1,
        STAFCommandParser::kValueRequired);

    // create needs

    fCreateParser.addOptionNeed("CREATE", "HANDLE");
    fCreateParser.addOptionNeed("CREATE", "NAME");

    // notify options

    fNotifyParser.addOption("STAF_NOTIFY", 1,
        STAFCommandParser::kValueNotAllowed);
    fNotifyParser.addOption("REGISTER", 1,
        STAFCommandParser::kValueNotAllowed);
    fNotifyParser.addOption("ONENDOFHANDLE", 1,
        STAFCommandParser::kValueRequired);
    fNotifyParser.addOption("MACHINE", 1,
        STAFCommandParser::kValueRequired);
    fNotifyParser.addOption("UUID", 1,
        STAFCommandParser::kValueRequired);
    fNotifyParser.addOption("SERVICE", 1,
        STAFCommandParser::kValueRequired);
    fNotifyParser.addOption("KEY", 1,
        STAFCommandParser::kValueRequired);

    // notify needs

    fNotifyParser.addOptionNeed("STAF_NOTIFY", "REGISTER");
    fNotifyParser.addOptionNeed("STAF_NOTIFY", "ONENDOFHANDLE");
    fNotifyParser.addOptionNeed("STAF_NOTIFY", "MACHINE");
    fNotifyParser.addOptionNeed("STAF_NOTIFY", "SERVICE");
    fNotifyParser.addOptionNeed("STAF_NOTIFY", "KEY");

    // list options

    fListParser.addOption("LIST",          1,
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("NOTIFICATIONS", 1,
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("HANDLE",        1,
        STAFCommandParser::kValueRequired);
    fListParser.addOption("MACHINE",        1,
        STAFCommandParser::kValueRequired);

    // these are for LIST HANDLES
    fListParser.addOption("HANDLES",        1,
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("NAME",           1,
        STAFCommandParser::kValueRequired);
    fListParser.addOption("PENDING",        1,
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("REGISTERED",     1,
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("INPROCESS",     1,
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("STATIC",     1,
        STAFCommandParser::kValueNotAllowed);

    // LIST groups
    fListParser.addOptionGroup("HANDLES NOTIFICATIONS", 0, 1);

    // list needs
    fListParser.addOptionNeed("PENDING REGISTERED INPROCESS STATIC", "HANDLES");
    fListParser.addOptionNeed("HANDLE MACHINE", "NOTIFICATIONS");

    // unregister options

    fUnregisterParser.addOption("UNREGISTER", 1,
        STAFCommandParser::kValueNotAllowed);

    fUnregisterParser.addOption("ONENDOFHANDLE", 1,
        STAFCommandParser::kValueRequired);

    fUnregisterParser.addOption("MACHINE", 1,
        STAFCommandParser::kValueRequired);

    fUnregisterParser.addOption("NOTIFYSERVICE", 1,
        STAFCommandParser::kValueRequired);

    fUnregisterParser.addOption("NOTIFYMACHINE", 1,
        STAFCommandParser::kValueRequired);

    // unregister needs

    fUnregisterParser.addOptionNeed("STAF_NOTIFY", "ONENDOFHANDLE");
    fUnregisterParser.addOptionNeed("STAF_NOTIFY", "MACHINE");
    fUnregisterParser.addOptionGroup("NOTIFYSERVICE NOTIFYMACHINE", 0, 1);

    // delete options

    fDeleteParser.addOption("DELETE",      1,
        STAFCommandParser::kValueNotAllowed);
    fDeleteParser.addOption("HANDLE",      1,
        STAFCommandParser::kValueRequired);

    // delete needs

    fDeleteParser.addOptionNeed("DELETE", "HANDLE");

    // authenticate options

    fAuthParser.addOption("AUTHENTICATE", 1,
        STAFCommandParser::kValueNotAllowed);
    fAuthParser.addOption("USER",         1,
        STAFCommandParser::kValueRequired);
    fAuthParser.addOption("CREDENTIALS",  1,
        STAFCommandParser::kValueRequired);
    fAuthParser.addOption("AUTHENTICATOR",  1,
        STAFCommandParser::kValueRequired);

    // authenticate needs

    fAuthParser.addOptionNeed("AUTHENTICATE", "USER");
    fAuthParser.addOptionNeed("USER", "CREDENTIALS");

    // un-authenticate options

    fUnAuthParser.addOption("UNAUTHENTICATE", 1,
        STAFCommandParser::kValueNotAllowed);

    // Construct map class for handle information used in QUERY HANDLE

    fQueryHandleClass = STAFMapClassDefinition::create(
        "STAF/Service/Handle/QueryHandle");
    fQueryHandleClass->addKey("handle", "Handle");
    fQueryHandleClass->addKey("state", "State");
    fQueryHandleClass->addKey("lastUsedTimestamp", "Last Used Date-Time");
    fQueryHandleClass->addKey("user", "User");

    // Construct map class for handle information used in LIST HANDLES

    fHandleInfoClass = STAFMapClassDefinition::create(
        "STAF/Service/Handle/HandleInfo");
    fHandleInfoClass->addKey("handle", "Handle");
    fHandleInfoClass->setKeyProperty("handle", "display-short-name", "H#");
    fHandleInfoClass->addKey("name", "Handle Name");
    fHandleInfoClass->setKeyProperty("name", "display-short-name", "Name");
    fHandleInfoClass->addKey("state", "State");
    fHandleInfoClass->addKey("lastUsedTimestamp", "Last Used Date-Time");

    // Construct map class for notifiee information used in LIST NOTIFICATIONS

    fNotificationClass = STAFMapClassDefinition::create(
        "STAF/Service/Handle/NotificationInfo");
 
    fNotificationClass->addKey("handle", "Handle");
    fNotificationClass->addKey("machine", "Machine");
    fNotificationClass->addKey("notifyService", "Notify Service");
}


STAFHandleService::~STAFHandleService()
{
    /* Do Nothing */
}


STAFString STAFHandleService::info(unsigned int) const
{
    return name() + ": Internal";
}

void STAFHandleService::handleRemoved(STAFHandle_t handle)
{
    if (debug) // XXX
        cout << "STAFHandleService handleRemoved handle=" << handle << "\n";

    STAFHandleManager::NotificationList notificationList =
        gHandleManagerPtr->getNotificationListCopy();

    STAFHandleManager::NotificationList::iterator iter;

    for (iter = notificationList.begin();
         iter != notificationList.end(); iter++)
    {
        STAFHandleManager::NotificationData notifiee(*iter);

        if (notifiee.handle == handle)
        {
            if (debug) // XXX
                cout << "Found a matching handle " << handle << "\n";

            STAFString request = "STAF_CALLBACK HANDLEDELETED HANDLE ";
            request += notifiee.handle;
            request += " MACHINE ";
            request += gMachinePtr->getImpl();
            request += " UUID ";
            request += notifiee.uuid;
            request += " KEY ";
            request += notifiee.key;

            if (debug) // XXX
                cout << notifiee.machine << " " << notifiee.notifyService << " " << request << "\n";
            STAFResultPtr result = 
                gSTAFProcHandlePtr->submit(notifiee.machine,
                                           notifiee.notifyService,
                                           request,
                                           kSTAFReqFireAndForget);
            if (debug) // XXX
                cout << "rc: " << result->rc << " result: " << result->result << "\n";
            gHandleManagerPtr->deleteNotification(handle,
                                                  notifiee.machine,
                                                  notifiee.uuid,
                                                  notifiee.notifyService);
        }
        else
        {
            if (debug) // XXX
                cout << "Found an unmatching handle " << handle << " " << notifiee.handle << "\n";
        }
    }
}


STAFServiceResult STAFHandleService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    STAFString action = requestInfo.fRequest.subWord(0, 1).lowerCase();

    if      (action == "query")        return handleQuery(requestInfo);
    else if (action == "list")         return handleList(requestInfo);
    else if (action == "staf_notify")  return handleNotify(requestInfo);
    else if (action == "create")       return handleCreate(requestInfo);
    else if (action == "delete")       return handleDelete(requestInfo);
    else if (action == "authenticate") return handleAuthenticate(requestInfo);
    else if (action == "unauthenticate")
    {
        return handleUnAuthenticate(requestInfo);
    }
    else if (action == "help")         return handleHelp(requestInfo);
    else return STAFServiceResult(kSTAFInvalidRequestString, "");
}


STAFServiceResult STAFHandleService::handleCreate(
    const STAFServiceRequest &requestInfo)
{
    // Verify that this request came from the local machine and that
    // the requesting machine/user has at least trust level 5

    IVALIDATE_LOCAL_TRUST(5, "CREATE");
    
    // Parse the request

    STAFCommandParseResultPtr parsedResult = fCreateParser.parse(
        requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer);
    }

    DEFINE_VAR_POOL_LIST(varPoolList, varPoolListSize, requestInfo);
    STAFString errorBuffer;
    STAFString name;
    STAFRC_t rc = RESOLVE_STRING_OPTION("NAME", name);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFHandle_t theHandle = 0;

    rc = gHandleManagerPtr->addAndGetStaticHandle(theHandle, name);

    if (rc) return rc;

    return STAFServiceResult(kSTAFOk, theHandle);
}


STAFServiceResult STAFHandleService::handleDelete(
    const STAFServiceRequest &requestInfo)
{
    // Verify that this request came from the local machine and that
    // the requesting machine/user has at least trust level 5

    IVALIDATE_LOCAL_TRUST(5, "DELETE");

    // Parse the request

    STAFCommandParseResultPtr parsedResult =
                              fDeleteParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer);
    }

    DEFINE_VAR_POOL_LIST(varPoolList, varPoolListSize, requestInfo);
    STAFString errorBuffer;
    STAFHandle_t theHandle = 0;
    STAFRC_t rc = RESOLVE_UINT_OPTION("HANDLE", theHandle);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    rc = gHandleManagerPtr->removeStaticHandle(theHandle);

    handleRemoved(theHandle);

    if (rc) return STAFServiceResult(rc, theHandle);

    return kSTAFOk;
}


STAFServiceResult STAFHandleService::handleQuery(
    const STAFServiceRequest &requestInfo)
{
    // Verify that the requesting machine/user has at least trust level 2

    IVALIDATE_TRUST(2, "QUERY");
    
    // Parse the request

    STAFCommandParseResultPtr parsedResult = fQueryParser.parse(
        requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer);
    }

    // Resolve the HANDLE option value

    DEFINE_VAR_POOL_LIST(varPoolList, varPoolListSize, requestInfo);
    STAFString errorBuffer;
    unsigned int handleNumber;
    
    STAFRC_t rc = RESOLVE_OPTIONAL_UINT_OPTION("HANDLE", handleNumber);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    // Get a copy of the handle list

    STAFHandleManager::HandleList handleList =
                                  gHandleManagerPtr->getHandleListCopy();
    STAFHandleManager::HandleList::iterator iter;
    
    // Create a marshalled map containing information about the handle

    STAFObjectPtr mc = STAFObject::createMarshallingContext();
    mc->setMapClassDefinition(fQueryHandleClass->reference());
    STAFObjectPtr handleMap = fQueryHandleClass->createInstance();

    for (iter = handleList.begin(); iter != handleList.end(); iter++)
    {
        STAFHandleManager::HandleData handleData = iter->second;
        if (handleData.handle == handleNumber)
        {
            STAFString handleState;

            if (handleData.state == STAFHandleManager::kPending)
            {
                handleState = "Pending";
            }
            else if ((handleData.state == STAFHandleManager::kRegistered)
                 ||  (handleData.state == STAFHandleManager::kPendingRegistered))
            {
                handleState = "Registered";
            }
            else if (handleData.state == STAFHandleManager::kInProcess)
            {
                handleState = "InProcess";
            }
            else if (handleData.state == STAFHandleManager::kStatic)
            {
                handleState = "Static";
            }
            else
            {
                return STAFServiceResult(kSTAFUnknownError, "Unknown Handle State");
            }

            handleMap->put("handle", STAFString(handleNumber));

            if (handleData.name.length() != 0)
                handleMap->put("name", handleData.name);

            handleMap->put("state", handleState);
            handleMap->put(
                "lastUsedTimestamp", handleData.lastUsedTimestamp.asString());
            handleMap->put(
                "user", STAFString(handleData.authenticator) +
                gSpecSeparator + STAFString(handleData.userIdentifier));
        }
    }

    mc->setRootObject(handleMap);

    return STAFServiceResult(kSTAFOk, mc->marshall());
}


STAFServiceResult STAFHandleService::handleAuthenticate(
    const STAFServiceRequest &requestInfo)
{
    // Verify that this request came from the local machine and that
    // the requesting machine/user has at least trust level 5

    IVALIDATE_LOCAL_TRUST(5, "AUTHENTICATE");

    // Parse the request

    STAFCommandParseResultPtr parsedResult =
                              fAuthParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer);
    }

    // Resolve any variables in USER, CREDENTIALS, and AUTHENTICATOR

    DEFINE_VAR_POOL_LIST(varPoolList, varPoolListSize, requestInfo);
    STAFString authenticator =  gServiceManagerPtr->getDefaultAuthenticator();
    STAFString userIdentifier;
    STAFString errorBuffer;

    STAFRC_t rc = RESOLVE_STRING_OPTION("USER", userIdentifier);

    if (!rc) rc = RESOLVE_OPTIONAL_STRING_OPTION("AUTHENTICATOR",
                                                 authenticator);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    if (authenticator == gNoneString)
    {
        return STAFServiceResult(kSTAFHandleAuthenticationDenied,
                                 "No Authenticators are registered");
    }

    STAFString credentials = parsedResult->optionValue("CREDENTIALS");

    // Issue an Authenticate request to the specified Authenticator service
    // passing in the credentials.  If the handle is already authenticated,
    // or if the authenticator specified is not registered or if authentication
    // fails, an error is returned.  If successful, authentication data is
    // returned in the result.

    STAFServiceResult serviceResult = gHandleManagerPtr->authenticate(
        requestInfo.fMachine, requestInfo.fHandle, authenticator,
        userIdentifier, STAFHandleManager::kCredentials, credentials);

    if (serviceResult.fRC != kSTAFOk) return serviceResult;

    // User was successfully authenticated

    STAFString authenticationData = serviceResult.fResult;

    // Cache the authentication information for the machine/handle/user/data
    // Can ignore any errors returned by cacheAuthenticationInfo() as an
    // error can only occur if the handle no longer exists, which should
    // never happen, but if it did, we wouldn't need to cache authinfo for it.

    gHandleManagerPtr->cacheAuthenticationInfo(requestInfo.fMachine,
                                               requestInfo.fHandle,
                                               authenticator,
                                               userIdentifier,
                                               authenticationData);

    return STAFServiceResult(kSTAFOk);
}


STAFServiceResult STAFHandleService::handleUnAuthenticate(
    const STAFServiceRequest &requestInfo)
{
    // Verify that this request came from the local machine and that
    // the requesting machine/user has at least trust level 5

    IVALIDATE_LOCAL_TRUST(5, "UNAUTHENTICATE");

    // Parse the request

    STAFCommandParseResultPtr parsedResult = fUnAuthParser.parse(
        requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer);
    }

    // Un-authenticate the handle

    return gHandleManagerPtr->unAuthenticate(requestInfo.fHandle);
}


STAFServiceResult STAFHandleService::handleList(
    const STAFServiceRequest &requestInfo)
{
    // Verify that the requesting machine/user has at least trust level 2

    IVALIDATE_TRUST(2, "LIST");

    // Parse the request

    STAFCommandParseResultPtr parsedResult = fListParser.parse(
        requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer);
    }

    STAFObjectPtr mc = STAFObject::createMarshallingContext();

    if (parsedResult->optionTimes("NOTIFICATIONS") != 0)
    {
        STAFHandleManager::NotificationList notificationList =
            gHandleManagerPtr->getNotificationListCopy();
        
        // Create a marshalled list of maps containing notification information

        mc->setMapClassDefinition(fNotificationClass->reference());
        STAFObjectPtr outputList = STAFObject::createList();
        
        STAFHandleManager::NotificationList::iterator iter;

        for (iter = notificationList.begin();
             iter != notificationList.end(); iter++)
        {
            STAFHandleManager::NotificationData notifiee(*iter);

            STAFObjectPtr notifyMap = fNotificationClass->createInstance();
            notifyMap->put("handle", STAFString(notifiee.handle));
            notifyMap->put("machine", STAFString(notifiee.machine));
            notifyMap->put("notifyService",
                           STAFString(notifiee.notifyService));
            outputList->append(notifyMap);
        }
        
        mc->setRootObject(outputList);
    }
    else //list handles
    {
        int getPending = 0;
        int getRegistered = 0;
        int getInProcess = 0;
        int getStatic = 0;

        if ((parsedResult->optionTimes("PENDING") == 0) &&
            (parsedResult->optionTimes("REGISTERED") == 0) &&
            (parsedResult->optionTimes("INPROCESS") == 0) &&
            (parsedResult->optionTimes("STATIC") == 0))
        {
            getPending = 0;
            getRegistered = 1;
            getInProcess = 1;
            getStatic = 1;
        }
        else
        {
            if (parsedResult->optionTimes("PENDING") != 0)
                getPending = 1;
            if (parsedResult->optionTimes("REGISTERED") != 0)
                getRegistered = 1;
            if (parsedResult->optionTimes("INPROCESS") != 0)
                getInProcess = 1;
            if (parsedResult->optionTimes("STATIC") != 0)
                getStatic = 1;
        }

        unsigned int getAll = parsedResult->optionTimes("NAME") == 0;

        // Resolve the NAME option value

        DEFINE_VAR_POOL_LIST(varPoolList, varPoolListSize, requestInfo);
        STAFString errorBuffer;
        STAFString name;
        STAFRC_t rc = RESOLVE_OPTIONAL_STRING_OPTION("NAME", name);

        if (rc) return STAFServiceResult(rc, errorBuffer);
        
        // Create the marshalled list of maps containing handle information

        mc->setMapClassDefinition(fHandleInfoClass->reference());
        STAFObjectPtr outputList = STAFObject::createList();
        
        // Get a copy of the handle list

        STAFHandleManager::HandleList handleList =
                                      gHandleManagerPtr->getHandleListCopy();
        STAFHandleManager::HandleList::iterator iter;

        for (iter = handleList.begin(); iter != handleList.end(); iter++)
        {
            STAFHandleManager::HandleData handleData = iter->second;

            if (((getAll != 0) || (handleData.name == name))
                && (((getPending == 1) && (handleData.state == STAFHandleManager::kPending))
                || ((getRegistered == 1) && ((handleData.state == STAFHandleManager::kRegistered)
                                || (handleData.state == STAFHandleManager::kPendingRegistered)))
                || ((getInProcess == 1) && (handleData.state == STAFHandleManager::kInProcess))
                || ((getStatic == 1) && (handleData.state == STAFHandleManager::kStatic))))
            {

                STAFObjectPtr handleMap = fHandleInfoClass->createInstance();
                handleMap->put("handle", STAFString(handleData.handle));

                if (handleData.state == STAFHandleManager::kPending)
                {
                    if (handleData.name.length() != 0)
                        handleMap->put("name", STAFString(handleData.name));

                    handleMap->put("state", STAFString("Pending"));
                }
                else if ((handleData.state == STAFHandleManager::kRegistered) ||
                         (handleData.state == STAFHandleManager::kPendingRegistered))
                {
                    handleMap->put("name", STAFString(handleData.name));
                    handleMap->put("state", STAFString("Registered"));
                }
                else if (handleData.state == STAFHandleManager::kInProcess)
                {
                    handleMap->put("name", STAFString(handleData.name));
                    handleMap->put("state", STAFString("InProcess"));
                }
                else if (handleData.state == STAFHandleManager::kStatic)
                {
                    handleMap->put("name", STAFString(handleData.name));
                    handleMap->put("state", STAFString("Static"));
                }
                else
                {
                    if (handleData.name.length() != 0)
                        handleMap->put("name", STAFString(handleData.name));
                    handleMap->put("state", STAFString("<Unknown>"));
                }

                handleMap->put("lastUsedTimestamp",
                               handleData.lastUsedTimestamp.asString());

                outputList->append(handleMap);
            }
        }
        
        mc->setRootObject(outputList);
    }

    return STAFServiceResult(kSTAFOk, mc->marshall());
}


STAFServiceResult STAFHandleService::handleNotify(
    const STAFServiceRequest &requestInfo)
{
    STAFCommandParseResultPtr parsedResult = fNotifyParser.parse(
        requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer);
    }

    DEFINE_VAR_POOL_LIST(varPoolList, varPoolListSize, requestInfo);
    STAFString errorBuffer;
    STAFHandle_t handle;
    STAFRC_t rc = RESOLVE_UINT_OPTION("ONENDOFHANDLE", handle);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    // Don't need to resolve these variables as STAF_NOTIFY request is not
    // documented and STAF doesn't need to pass any variables in these options

    STAFString machine = parsedResult->optionValue("MACHINE");
    STAFString service = parsedResult->optionValue("SERVICE");
    STAFString key     = parsedResult->optionValue("KEY");
    
    STAFString uuid;
    
    // UUID is optional.  If not specified, use MISC WHOAREYOU.
    
    if (parsedResult->optionTimes("UUID") == 0)
    {
        STAFResultPtr result = 
            gSTAFProcHandlePtr->submit(machine, "MISC", "WHOAREYOU");

        if (result->rc == kSTAFOk)
        {
            STAFObjectPtr mc = STAFObject::unmarshall(result->result);
            STAFObjectPtr poolMap = mc->getRootObject();
            uuid = poolMap->get("instanceUUID")->asString();
        }
        else
        {
            uuid = "N/A";
        }
    }
    else
    {
        uuid = parsedResult->optionValue("UUID");
    }

    return gHandleManagerPtr->addNotification(handle, machine, uuid, service,
                                               key);
}


STAFServiceResult STAFHandleService::handleHelp(
    const STAFServiceRequest &requestInfo)
{
    // Verify that the requesting machine/user has at least trust level 1

    IVALIDATE_TRUST(1, "HELP");

    STAFString result("HANDLE service help");

    result += *gLineSeparatorPtr + *gLineSeparatorPtr;
    result += "CREATE HANDLE NAME <Handle Name>" + *gLineSeparatorPtr +
              *gLineSeparatorPtr;
    result += "DELETE HANDLE <Number>" + *gLineSeparatorPtr +
              *gLineSeparatorPtr;
    result += "QUERY HANDLE <Handle>" + *gLineSeparatorPtr +
              *gLineSeparatorPtr;
    result += "LIST [HANDLES [NAME <Handle Name>] [PENDING] [REGISTERED] [INPROCESS] [STATIC]]" +
               *gLineSeparatorPtr + *gLineSeparatorPtr;
    result += "LIST NOTIFICATIONS [HANDLE <Handle> | MACHINE <Machine>]" +
               *gLineSeparatorPtr + *gLineSeparatorPtr;
    result += "AUTHENTICATE USER <User Identifier> CREDENTIALS <Credentials>" +
              *gLineSeparatorPtr +
              "             [AUTHENTICATOR <Authenticator Name>]" +
              *gLineSeparatorPtr + *gLineSeparatorPtr;
    result += "UNAUTHENTICATE" + *gLineSeparatorPtr + *gLineSeparatorPtr;
    result += "STAF_NOTIFY REGISTER ONENDOFHANDLE <Handle> MACHINE <Machine>" +
              *gLineSeparatorPtr;
    result += "            KEY <Key> SERVICE <Service>" +
              *gLineSeparatorPtr + *gLineSeparatorPtr;
    result += "HELP" + *gLineSeparatorPtr;

    return STAFServiceResult(kSTAFOk, result);
}
